from datetime import timedelta
import numpy as np
import pyproj
from bisect import bisect_left
from abc import abstractmethod
import logging
logger = logging.getLogger(__name__)

from opendrift.timer import Timeable
from opendrift.errors import OutsideSpatialCoverageError, OutsideTemporalCoverageError, VariableNotCoveredError
from .consts import standard_names, vector_pairs_xy


class ReaderDomain(Timeable):
    """
    Projection, spatial and temporal domain of reader.
    """
    name = None

    proj4 = None
    proj = None

    lon = None
    lat = None
    xmin = None
    xmax = None
    ymin = None
    ymax = None
    zmin = -np.inf
    zmax = np.inf
    delta_x = None
    delta_y = None
    shape = None

    ## Temporal
    start_time = None
    end_time = None
    time_step = None
    times = None

    """Setting this to `True` overrides temporal and spatial bounds checks.
    Also useful for readers that are constant and do not have a temporal
    dimension."""
    always_valid = False

    def center(self):
        """
        Returns center of reader (in lon, lat)
        """

        if any(xx is None for xx in [self.xmin, self.xmax, self.ymin, self.ymax]):
            return None, None

        x = self.xmin + (self.xmax - self.xmin) / 2
        y = self.ymin + (self.ymax - self.ymin) / 2
        return self.xy2lonlat(x, y)

    def rotate_vectors(self, reader_x, reader_y, u_component, v_component,
                       proj_from, proj_to):
        """Rotate vectors from one crs to another."""

        if type(proj_from) is str:
            proj_from = pyproj.Proj(proj_from)
        if type(proj_from) is not pyproj.Proj:
            proj_from = pyproj.Proj('+proj=latlong +R=6370997.0 +ellps=WGS84')
            reader_x, reader_y = self.xy2lonlat(reader_x, reader_y)
        if type(proj_to) is str:
            proj_to = pyproj.Proj(proj_to)

        if proj_from.crs.is_geographic:
            delta_y = .1  # 0.1 degree northwards
        else:
            delta_y = 10  # 10 m along y-axis

        transformer = pyproj.Transformer.from_proj(proj_from, proj_to)
        x2, y2 = transformer.transform(reader_x, reader_y)
        x2_delta, y2_delta = transformer.transform(reader_x,
                                                   reader_y + delta_y)

        if proj_to.crs.is_geographic:
            geod = pyproj.Geod(ellps='WGS84')
            rot_angle_vectors_rad = np.radians(
                geod.inv(x2, y2, x2_delta, y2_delta)[0])
        else:
            rot_angle_vectors_rad = np.arctan2(x2_delta - x2, y2_delta - y2)
        logger.debug('Rotating vectors between %s and %s degrees.' %
                     (np.degrees(rot_angle_vectors_rad).min(),
                      np.degrees(rot_angle_vectors_rad).max()))
        rot_angle_rad = -rot_angle_vectors_rad
        u_rot = (u_component * np.cos(rot_angle_rad) -
                 v_component * np.sin(rot_angle_rad))
        v_rot = (u_component * np.sin(rot_angle_rad) +
                 v_component * np.cos(rot_angle_rad))

        return u_rot, v_rot

    def xy2lonlat(self, x, y):
        """Calculate x,y in own projection from given lon,lat (scalars/arrays).
        """
        x = np.atleast_1d(x)
        y = np.atleast_1d(y)
        if self.proj.crs.is_geographic:
            if 'ob_tran' in str(self.proj4):
                logger.debug('NB: Converting degrees to radians ' +
                             'due to ob_tran srs')
                x = np.radians(np.array(x))
                y = np.radians(np.array(y))
                return self.proj(x, y, inverse=True)
            else:
                return x, y
        else:
            return self.proj(x, y, inverse=True)

    def lonlat2xy(self, lon, lat):
        """
        Calculate lon,lat from given x,y (scalars/arrays) in own projection.
        """
        lon = np.atleast_1d(lon)
        lat = np.atleast_1d(lat)
        if 'ob_tran' in str(self.proj4):
            x, y = self.proj(lon, lat, inverse=False)
            return np.degrees(x), np.degrees(y)
        elif self.proj.crs.is_geographic:
            return lon, lat
        else:
            x, y = self.proj(lon, lat, inverse=False)
            return x, y

    def y_azimuth(self, lon, lat):
        """Calculate azimuth orientation of the y-axis of the reader CRS."""
        x0, y0 = self.lonlat2xy(lon, lat)
        distance = 1000.0  # Create points 1 km away to determine azimuth
        lon_2, lat_2 = self.xy2lonlat(x0, y0 + distance)
        geod = pyproj.Geod(ellps='WGS84')  # Define an ellipsoid
        y_az = geod.inv(lon_2, lat_2, lon, lat, radians=False)
        return y_az[0]

    def pixel_size(self):
        # Find typical pixel size (e.g. for calculating size of buffer)
        if self.delta_x is not None:
            pixelsize = self.delta_x
            if self.proj.crs.is_geographic is True or \
                    ('ob_tran' in self.proj4) or \
                    ('longlat' in self.proj4) or \
                    ('latlon' in self.proj4):
                pixelsize = pixelsize * 111000  # deg to meters
        else:
            pixelsize = None  # Pixel size not defined
        return pixelsize

    def _coverage_unit_(self):
        return "degrees"

    def __repr__(self):
        """String representation of the current reader."""
        outStr = '===========================\n'
        outStr += 'Reader: ' + self.name + '\n'
        outStr += 'Projection: \n  ' + self.proj4 + '\n'
        outStr += 'Coverage: [%s]\n' % self._coverage_unit_()
        shape = self.shape
        if shape is None:
            outStr += '  xmin: %f   xmax: %f\n' % (self.xmin, self.xmax)
            outStr += '  ymin: %f   ymax: %f\n' % (self.ymin, self.ymax)
        else:
            outStr += '  xmin: %f   xmax: %f   step: %g   numx: %i\n' % \
                (self.xmin, self.xmax, self.delta_x or 0, shape[0])
            outStr += '  ymin: %f   ymax: %f   step: %g   numy: %i\n' % \
                (self.ymin, self.ymax, self.delta_y or 0, shape[1])
        corners = self.xy2lonlat([self.xmin, self.xmin, self.xmax, self.xmax],
                                 [self.ymax, self.ymin, self.ymax, self.ymin])
        outStr += '  Corners (lon, lat):\n'
        outStr += '    (%6.2f, %6.2f)  (%6.2f, %6.2f)\n' % \
            (corners[0][0],
             corners[1][0],
             corners[0][2],
             corners[1][2])
        outStr += '    (%6.2f, %6.2f)  (%6.2f, %6.2f)\n' % \
            (corners[0][1],
             corners[1][1],
             corners[0][3],
             corners[1][3])
        if hasattr(self, 'z'):
            np.set_printoptions(suppress=True)
            outStr += 'Vertical levels [m]: \n  ' + str(self.z) + '\n'
        elif hasattr(self, 'sigma'):
            outStr += 'Vertical levels [sigma]: \n  ' + str(self.sigma) + '\n'
        else:
            outStr += 'Vertical levels [m]: \n  Not specified\n'
        outStr += 'Available time range:\n'
        outStr += '  start: ' + str(self.start_time) + \
                  '   end: ' + str(self.end_time) + \
                  '   step: ' + str(self.time_step) + '\n'
        if self.start_time is not None and self.time_step is not None:
            outStr += '    %i times (%i missing)\n' % (
                self.expected_time_steps, self.missing_time_steps)
        if hasattr(self, 'realizations') and self.realizations is not None:
            outStr += 'Variables (%i ensemble members):\n' % len(
                self.realizations)
        else:
            outStr += 'Variables:\n'
        for variable in self.variables:
            if variable in self.derived_variables:
                outStr += '  ' + variable + ' - derived from ' + \
                    str(self.derived_variables[variable]) + '\n'
            else:
                outStr += '  ' + variable + '\n'
        outStr += '===========================\n'
        outStr += self.performance()

        return outStr

    def covers_positions_xy(self, x, y, z=0):
        """
        Return indices of input points covered by reader.

        Arguments in native projection of reader.
        """
        x = np.atleast_1d(x)
        y = np.atleast_1d(y)
        z = np.atleast_1d(z)

        if self.global_coverage():
            # We need only check north-south and z coverage
            indices = np.where((y >= self.ymin) & (y <= self.ymax)
                               & (z >= self.zmin) & (z <= self.zmax))[0]
        else:
            xx = x

            if self.proj.crs.is_geographic:
                xx = self.modulate_longitude(x)

            indices = np.where((xx >= self.xmin) & (xx <= self.xmax)
                                & (y >= self.ymin) & (y <= self.ymax)
                                & (z >= self.zmin) & (z <= self.zmax))[0]

        try:
            return indices, x[indices], y[indices]
        except Exception as ex:
            logger.exception(ex)
            return indices, x, y

    def modulate_longitude(self, lons):
        """
        Modulate the input longitude to the domain supported by the reader.
        """

        yy = np.array([self.ymin, self.ymax, self.ymax, self.ymin])
        xx = np.array([self.xmin, self.xmin, self.xmax, self.xmax])

        exlons, _lats = self.xy2lonlat(xx, yy)

        if np.min(exlons) < 0:
            # Reader domain is from -180 to 180 or somewhere in between.
            assert np.max(exlons) <= 180.1

            lons = np.mod(lons+180, 360) - 180
        else:
            # Reader domain is from 0 to 360 or somewhere in between.
            assert np.max(exlons) <= 360.1

            lons = np.mod(lons, 360)

        return lons

    def covers_positions(self, lon, lat, z=0):
        """Return indices of input points covered by reader."""

        x, y = self.lonlat2xy(lon, lat)

        return self.covers_positions_xy(x, y, z)

    def global_coverage(self):
        """Return True if global coverage east-west"""

        dx = self.delta_x if self.delta_x is not None else 0
        if self.proj.crs.is_geographic is True:
            if (self.xmin - 2*dx <= 0) and (self.xmax + 2*dx >=
                                                    360):
                return True  # Global 0 to 360
            if (self.xmin - 2*dx <= -180) and (self.xmax + 2*dx
                                                       >= 180):
                return True  # Global -180 to 180

        return False

    def domain_grid(self, npoints=1000):
        """Return arrays of lon,lat points spread evenly over reader domain."""
        numx = np.floor(np.sqrt(npoints))
        numy = np.floor(np.sqrt(npoints))
        x = np.linspace(self.xmin, self.xmax - self.delta_x, numx)
        y = np.linspace(self.ymin, self.ymax - self.delta_y, numy)
        x, y = np.meshgrid(x, y)
        lons, lats = self.xy2lonlat(x, y)
        return lons, lats

    def coverage_string(self):
        """Coverage of reader to be reported as string for debug output"""
        corners = self.xy2lonlat([self.xmin, self.xmin, self.xmax, self.xmax],
                                 [self.ymax, self.ymin, self.ymax, self.ymin])
        return '%.2f-%.2fE, %.2f-%.2fN' % (np.min(
            corners[0]), np.max(corners[0]), np.min(
                corners[1]), np.max(corners[1]))

    def check_arguments(self, variables, time, x, y, z):
        """Check validity of arguments input to method get_variables.

        Checks that requested positions and time are within coverage of
        this reader, and that it can provide the requested variable(s).
        Returns the input arguments, possibly modified/corrected (below)

        Arguments:
            See function get_variables for definition.

        Returns:
            variables: same as input, but converted to list if given as string.

            time: same as input, or start_time of reader if given as None.

            x, y, z: same as input, but converted to ndarrays
            if given as scalars.

            outside: boolean array which is True for any particles outside
                the spatial domain covered by this reader.

        Raises:

            ValueError:

            - if requested time is outside coverage of reader.

            - if all requested positions are outside coverage of reader.
        """

        # Check time
        if time is None:
            time = self.start_time  # Use first timestep, if not given

        # Convert variables to list and x,y to ndarrays
        if isinstance(variables, str):
            variables = [variables]
        x = np.atleast_1d(x)
        y = np.atleast_1d(y)
        if z is not None:
            z = np.asarray(z)

        for variable in variables:
            if variable not in self.variables:
                raise VariableNotCoveredError('Variable not available: ' + variable +
                                 '\nAvailable parameters are: ' +
                                 str(self.variables))
        if (self.start_time is not None
                and time < self.start_time) and self.always_valid is False:
            raise OutsideTemporalCoverageError('Requested time (%s) is before first available '
                             'time (%s) of %s' % (time, self.start_time, self.name))
        if (self.end_time is not None
                and time > self.end_time) and self.always_valid is False:
            raise OutsideTemporalCoverageError('Requested time (%s) is after last available '
                             'time (%s) of %s' % (time, self.end_time, self.name))
        if self.global_coverage():
            outside = np.where(~np.isfinite(x + y) | (y < self.ymin)
                               | (y > self.ymax))[0]
        else:
            outside = np.where(~np.isfinite(x + y) | (x < self.xmin)
                               | (x > self.xmax) | (y < self.ymin)
                               | (y > self.ymax))[0]
        if np.size(outside) == np.size(x):
            lon, lat = self.xy2lonlat(x, y)
            raise OutsideSpatialCoverageError(('Argcheck: all %s particles (%.2f-%.2fE, ' +
                    '%.2f-%.2fN) are outside domain of %s (%s)') %
                    (len(lon), lon.min(), lon.max(), lat.min(),
                     lat.max(), self.name, self.coverage_string()))

        return variables, time, x, y, z, outside

    def covers_time(self, time):
        if self.always_valid is True:
            return True
        if self.start_time is None:
            return True  # No time limitations of reader
        if (time < self.start_time) or (time > self.end_time):
            return False
        else:
            return True

    def nearest_time(self, time):
        """Return nearest times before and after the requested time.

        Returns:
            nearest_time: datetime
            time_before: datetime
            time_after: datetime
            indx_nearest: int
            indx_before: int
            indx_after: int
        """
        if self.start_time == self.end_time:
            return self.start_time, self.start_time, None, 0, 0, 0
        if self.start_time is None:
            return None, None, None, None, None, None
        if self.times is not None:  # Time as array, possibly with holes
            indx_before = np.max((0, bisect_left(self.times, time) - 1))
            if len(self.times) > 1 and self.times[indx_before + 1] == time:
                # Correction needed when requested time exists in times
                indx_before = indx_before + 1
            time_before = self.times[indx_before]
            indx_after = np.minimum(indx_before + 1,
                                    len(self.times) - 1)  # At the end
            time_after = self.times[indx_after]
            if (time - time_before) < (time_after - time):
                indx_nearest = indx_before
            else:
                indx_nearest = indx_after
            nearest_time = self.times[indx_nearest]
        else:  # Time step is constant (no holes)
            if self.time_step is None:
                return None, None, None, None, None, None
            indx = float((time - self.start_time).total_seconds()) / \
                float(self.time_step.total_seconds())
            indx_nearest = int(round(indx))
            nearest_time = self.start_time + indx_nearest * self.time_step
            indx_before = int(np.floor(indx))
            time_before = self.start_time + indx_before * self.time_step
            indx_after = int(np.ceil(indx))
            time_after = self.start_time + indx_after * self.time_step
        return nearest_time, time_before, time_after,\
            indx_nearest, indx_before, indx_after

################################################################
# Methods to derive environment variables from others available
################################################################

def land_binary_mask_from_ocean_depth(env):
    env['land_binary_mask'] = np.float32(env['sea_floor_depth_below_sea_level'] <= 0)

def wind_from_speed_and_direction(env):
    if 'wind_from_direction' in env:
        wfd = env['wind_from_direction']
    else:
        wfd = -env['wind_to_direction']
    north_wind = -env['wind_speed']*np.cos(np.radians(wfd))
    east_wind = -env['wind_speed']*np.sin(np.radians(wfd))
    env['x_wind'] = east_wind
    env['y_wind'] = north_wind
    # Rotating might be necessary generally
    #x,y = np.meshgrid(env['x'], env['y'])
    #env['x_wind'], env['y_wind'] = self.rotate_vectors(
    #    x, y,
    #    east_wind, north_wind,
    #    None, self.proj)

################################################################

class Variables(ReaderDomain):
    """
    Handles reading and interpolation of variables.
    """
    variables = None
    derived_variables = None
    name = None

    buffer = 0

    def __init__(self):
        if self.derived_variables is None:
            self.derived_variables = {}
        if self.variables is None:
            self.variables = []

        # Deriving environment variables from other available variables
        self.environment_mappings = {
            'wind_from_speed_and_direction': {
                'input': ['wind_speed', 'wind_from_direction'],
                'output': ['x_wind', 'y_wind'],
                'method': wind_from_speed_and_direction,
                #lambda reader, env: reader.wind_from_speed_and_direction(env)},
                'active': True},
            'wind_from_speed_and_direction_to': {
                'input': ['wind_speed', 'wind_to_direction'],
                'output': ['x_wind', 'y_wind'],
                'method': wind_from_speed_and_direction,
                #lambda reader, env: reader.wind_from_speed_and_direction(env)},
                'active': True},
            'land_binary_mask_from_ocean_depth': {
                'input': ['sea_floor_depth_below_sea_level'],
                'output': ['land_binary_mask'],
                'method': land_binary_mask_from_ocean_depth,
                'active': False}
            }

        super().__init__()

    def activate_environment_mapping(self, mapping_name):
        if mapping_name not in self.environment_mappings:
            raise ValueError('Available environment mappings: ' + str(self.environment_mappings))

        em = self.environment_mappings[mapping_name]
        em['active'] = True
        if not all(item in em['output'] for item in self.variables) and \
                    all(item in self.variables for item in em['input']):
            for v in em['output']:
                logger.debug('Adding variable mapping: %s -> %s' % (em['input'], v))
                self.variables.append(v)
                self.derived_variables[v] = em['input']

    def __calculate_derived_environment_variables__(self, env):
        for m in self.environment_mappings:
            em = self.environment_mappings[m]
            if em['active'] is False:
                continue
            if not all(item in em['output'] for item in self.variables) and \
                    all(item in self.variables for item in em['input']):
                for v in em['output']:
                    logger.debug('Calculating variable mapping: %s -> %s' % (em['input'], v))
                    method = lambda env: em['method'](env)
                    method(env)

    def set_buffer_size(self, max_speed, time_coverage=None):
        '''
        Adjust buffer to minimise data block size needed to cover elements.

        The buffer size is calculated from the maximum anticpated speed.
        Seeding over a large area or over longer time can easily cause
        particles to be located outside the block. This is not critical, but
        causes interpolation to be one-sided in time for the relevant
        particles.

        Args:

            max_speed (m/s): the maximum speed anticipated for particles.
            time_coverage (timedelta): the time span to cover

        '''
        self.buffer = 0
        pixelsize = self.pixel_size()
        if pixelsize is not None:
            if self.time_step is not None:
                time_step_seconds = self.time_step.total_seconds()
            else:
                if time_coverage is None:
                    logger.warning('Assuming time step of 1 hour for ' + self.name)
                    time_step_seconds = 3600  # 1 hour if not given
                else:
                    time_step_seconds = time_coverage.total_seconds()
            self.buffer = int(
                np.ceil(max_speed * time_step_seconds / pixelsize)) + 2
            logger.debug('Setting buffer size %i for reader %s, assuming '
                         'a maximum average speed of %g m/s and time span of %s' %
                         (self.buffer, self.name, max_speed, timedelta(seconds=time_step_seconds)))

    def __check_env_coordinates__(self, env):
        """
        Make sure x and y are floats (and not e.g. int64)
        """
        if 'x' in env.keys():
            env['x'] = np.array(env['x'], dtype=np.float32)
            env['y'] = np.array(env['y'], dtype=np.float32)

    @staticmethod
    def __check_variable_array__(name, variable):
        """
        Ensure arrays are not masked arrays and that values are within valid ranges.
        """

        # Convert any masked arrays to NumPy arrays
        if isinstance(variable, np.ma.MaskedArray):
            variable = variable.filled(np.nan)

        # Mask values outside valid_min, valid_max (self.standard_names)
        if name in standard_names.keys():
            logger.debug("Checking %s for invalid values" % name)
            if isinstance(variable, list):
                logger.debug(
                    'Min-max checking for ensemble data (%i members)' %
                    (len(variable)))
                # Recursive
                return [
                    Variables.__check_variable_array__(name, v)
                    for v in variable
                ]
            # with np.errstate(invalid='ignore'):
            invalid_indices = np.logical_and(
                np.isfinite(variable),
                np.logical_or(variable < standard_names[name]['valid_min'],
                              variable > standard_names[name]['valid_max']))
            if np.sum(invalid_indices) > 0:
                invalid_values = variable[invalid_indices]
                logger.warning(
                    'Invalid values (%s to %s) found for %s, replacing with NaN'
                    % (invalid_values.min(), invalid_values.max(), name))
                logger.warning('(allowed range: [%s, %s])' %
                               (standard_names[name]['valid_min'],
                                standard_names[name]['valid_max']))
                variable[invalid_indices] = np.nan

        return variable

    def __check_env_arrays__(self, env):
        """
        Ensure arrays are not masked arrays and that values are within valid
        ranges.

        .. seealso::

            Disabled in `StructuredReader` because variables are valided before
            entered into interpolator:

            :meth:`.structured.StructuredReader.__check_env_arrays__`
        """
        variables = [
            var for var in env.keys() if var not in ['x', 'y', 'time']
        ]

        for variable in variables:
            env[variable] = self.__check_variable_array__(
                variable, env[variable])

        return env

    @abstractmethod
    def _get_variables_interpolated_(self, variables, profiles, profiles_depth,
                                     time, reader_x, reader_y, z):
        """
        This method _must_ be implemented by every reader. Usually by
        subclassing one of the reader types (e.g.
        :class:`structured.StructuredReader`).

        Arguments are in _native projection_ of reader.

        .. seealso:

            * :meth:`get_variables_interpolated_xy`.
            * :meth:`get_variables_interpolated`.
        """
        pass

    def get_variables_interpolated_xy(self,
                                      variables,
                                      profiles=None,
                                      profiles_depth=None,
                                      time=None,
                                      x=None,
                                      y=None,
                                      z=None,
                                      rotate_to_proj=None):
        """
        Get variables in native projection of reader.

        .. seealso::

            * :meth:`get_variables_interpolated`.
            * :meth:`_get_variables_interpolated_`.
        """
        self.timer_start('total')
        # Raise error if time not not within coverage of reader
        if not self.covers_time(time):
            raise OutsideTemporalCoverageError('%s is outside time coverage (%s - %s) of %s' %
                             (time, self.start_time, self.end_time, self.name))

        self.timer_start('preparing')

        x = np.atleast_1d(x)
        numx = len(x)  # To check later if all points were covered

        y = np.atleast_1d(y)
        z = np.atleast_1d(z) if z is not None else np.zeros((1, ))

        assert numx == len(y), "x, y, and z must have the same length"
        assert len(z) == 1 or numx == len(
            z), "x, y, and z must have the same length"

        ind_covered, xx, yy = self.covers_positions_xy(x, y, z)
        if len(ind_covered) == 0:
            lon, lat = self.xy2lonlat(x, y)
            raise OutsideSpatialCoverageError(
                ('All %s particles (%.2f-%.2fE, %.2f-%.2fN) ' +
                 'are outside domain of %s (%s)') %
                 (len(x), lon.min(), lon.max(), lat.min(),
                 lat.max(), self.name, self.coverage_string()))
        x = xx
        y = yy

        self.timer_end('preparing')

        # Make copy of z to avoid modifying original array
        if len(z) == 1 and len(x) > 1:
            z = z.copy() * np.ones(x.shape)
        z = z.copy()[ind_covered]

        logger.debug('Fetching variables from ' + self.name)
        self.timer_start('reading')

        # Filter derived variables
        derived = []
        derived_input = []

        for var in variables:
            if var in self.derived_variables:
                logger.debug("Scheduling %s to be derived from %s" %
                             (var, self.derived_variables[var]))
                derived_input.extend(self.derived_variables[var])
                derived.append(var)

        for v in derived:
            variables.remove(v)
        variables.extend(list(set(derived_input)))

        env, env_profiles = self._get_variables_interpolated_(
            variables, profiles, profiles_depth, time, x, y, z)

        # Calculate derived variables
        if len(derived) > 0:
            self.__calculate_derived_environment_variables__(env)
            variables.extend(derived)

        for e in [env, env_profiles]:
            if e is not None:
                self.__check_env_coordinates__(e)
                self.__check_env_arrays__(e)

        self.timer_end('reading')

        # Rotating vectors fields
        if rotate_to_proj is not None:
            if self.proj.crs.is_geographic:
                logger.debug('Reader projection is latlon - '
                             'rotation of vectors is not needed.')
            else:
                vector_pairs = []
                for var in variables:
                    for vector_pair in vector_pairs_xy:
                        if var in vector_pair:
                            counterpart = list(set(vector_pair) -
                                               set([var]))[0]
                            if counterpart in variables:
                                vector_pairs.append(vector_pair)
                            else:
                                logger.warning(
                                    'Missing component of vector pair, cannot rotate:'
                                    + counterpart)
                # Extract unique vector pairs
                vector_pairs = [
                    list(x) for x in set(tuple(x) for x in vector_pairs)
                ]

                if len(vector_pairs) > 0:
                    self.timer_start('rotating vectors')
                    for vector_pair in vector_pairs:
                        env[vector_pair[0]], env[vector_pair[1]] = \
                            self.rotate_vectors(x, y,
                                                env[vector_pair[0]],
                                                env[vector_pair[1]],
                                                self.proj, rotate_to_proj)
                        if profiles is not None and vector_pair[0] in profiles:
                            env_profiles[vector_pair[0]], env_profiles[vector_pair[1]] = \
                                    self.rotate_vectors (x, y,
                                            env_profiles[vector_pair[0]],
                                            env_profiles[vector_pair[1]],
                                            self.proj,
                                            rotate_to_proj)

                    self.timer_end('rotating vectors')

        # Masking non-covered pixels
        self.timer_start('masking')
        if len(ind_covered) != numx:
            logger.debug('Masking %i elements outside coverage' %
                         (numx - len(ind_covered)))
            for var in variables:
                tmp = np.nan * np.ones(numx)
                tmp[ind_covered] = env[var].copy()
                env[var] = np.ma.masked_invalid(tmp)
                # Filling also in missing columns
                # for env_profiles outside coverage
                if env_profiles is not None and var in env_profiles.keys():
                    tmp = np.nan * np.ones((env_profiles[var].shape[0], numx))
                    tmp[:, ind_covered] = env_profiles[var].copy()
                    env_profiles[var] = np.ma.masked_invalid(tmp)

        self.timer_end('masking')
        self.timer_end('total')

        return env, env_profiles

    def get_variables_interpolated(self,
                                   variables,
                                   profiles=None,
                                   profiles_depth=None,
                                   time=None,
                                   lon=None,
                                   lat=None,
                                   z=None,
                                   rotate_to_proj=None):
        """
        `get_variables_interpolated` is the main interface to
        :class:`opendrift.basemodel.OpenDriftSimulation`, and is responsible
        for returning variables at the correct positions.

        Readers should implement :meth:`_get_variables_interpolated_`.

        Arguments:
            variables: string, or list of strings (standard_name) of
                requested variables. These must be provided by reader.

            profiles: List of variable names that should be returned for the range in `profiles_depth`.

            profiles_depth: A range [z-start, z-end] for which to return values for profile-variables. The exact z-depth are given by the reader and returned as `z` variable in `env_profiles`.

            time: datetime or None, time at which data are requested.
                Can be None (default) if reader/variable has no time
                dimension (e.g. climatology or landmask).

            lon: longitude, 1d array.

            lat: latitude, 1d array, same length as lon.

            z: float or ndarray; vertical position (in meters, positive up)
                of requested points. either scalar or same length as lon, lat.
                default: 0 m (unless otherwise documented by reader)

            block: bool, see return below

            rotate_to_proj: N/A

          Returns:

            (env, env_profiles)

            Interpolated variables at x, y and z. `env` contains values at a fixed depth (`z`), while `env_profiles` contains depth-profiles in the range `profile_depth` for the variables listed in `profiles` for each element (in `x`, `y`). The exact depth is determined by the reader and specified in
            `env_profiles['z']`. Thus variables in `env_profiles` are not interpolated in z-direction.

        .. seealso::

            :meth:`get_variables_interpolated_xy`.

        """
        assert set(variables).issubset(self.variables), f"{variables} is not subset of {self.variables}"

        lon = self.modulate_longitude(lon)
        x, y = self.lonlat2xy(lon, lat)

        env, env_profiles = self.get_variables_interpolated_xy(
            variables, profiles, profiles_depth, time, x, y, z, rotate_to_proj)

        return env, env_profiles



class NotCoveredError(Exception):
    pass

class OutsideSpatialCoverageError(NotCoveredError):
    pass

class OutsideTemporalCoverageError(NotCoveredError):
    pass

class VariableNotCoveredError(NotCoveredError):
    pass



# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2015, Knut-Frode Dagestad, MET Norway

from datetime import datetime
import logging
logger = logging.getLogger(__name__)

import numpy as np
try:
    import pygrib
except:
    raise ImportError('PyGrib library is needed for GRIB files: '
                      'https://jswhit.github.io/pygrib/docs/index.html')

from opendrift.readers.basereader import BaseReader, StructuredReader

# Hardcoded "GRIB-tables" for now.
grib_variable_mapping = {
    'enmi': {  # MET Norway (enmi) GRIB codes (marsParam)
        '33.3': 'x_wind',
        '34.3': 'y_wind',
        '49.3': 'x_sea_water_velocity',
        '50.3': 'y_sea_water_velocity',
        '100.3': 'sea_surface_wave_significant_height',
        '231.3':
            'sea_surface_wave_period_at_variance_spectral_density_maximum',
        '232.3':
            'sea_surface_wave_mean_period_from_variance_spectral_density_second_frequency_moment',
        '247.3': 'sea_surface_wave_stokes_drift_x_velocity',
        '248.3': 'sea_surface_wave_stokes_drift_y_velocity'},
    'ecmf': {  # ECMWF GRIB codes
        '165.128': 'x_wind',
        '166.128': 'y_wind',
        '215.140': 'sea_surface_wave_stokes_drift_x_velocity',
        '216.140': 'sea_surface_wave_stokes_drift_y_velocity',
        '229.140': 'sea_surface_wave_significant_height',
        '231.140':
            'sea_surface_wave_period_at_variance_spectral_density_maximum',
        '232.140':
            'sea_surface_wave_mean_period_from_variance_spectral_density_second_frequency_moment'
        },
    'kwbc': {  #
        '33.2': 'x_wind',
        '34.2': 'y_wind',
        '49.2': 'x_sea_water_velocity',
        '50.2': 'y_sea_water_velocity'}
     }


class Reader(BaseReader, StructuredReader):

    def __init__(self, filename=None, name=None):

        if filename is None:
            raise ValueError('Need filename as argument to constructor')

        if name is None:
            self.name = filename
        else:
            self.name = name

        try:
            # Open file, check that everything is ok
            logger.info('Opening dataset: ' + filename)
            self.grib = pygrib.open(filename)
        except:
            raise ValueError('Could not open ' + filename +
                             ' with pygrib library')

        ####################
        # Scan file
        ####################
        levels = []
        centre = []
        projs = []
        times = []
        marsParams = []
        for m in self.grib:
            levels.append(m.level)
            centre.append(m.centre)
            projs.append(m.projparams['proj'])
            s = '%s%04d' % (m.validityDate, m.validityTime)
            times.append(datetime.strptime(s, '%Y%m%d%H%M'))
            marsParams.append(m.marsParam)

        ################
        # Projection
        ################
        projs = list(set(projs))
        if len(projs) > 1:
            raise ValueError('File with data in several projections is not '
                             'supported: ' + str(projs))
        if projs[0] == 'cyl':
            self.proj4 = '+proj=latlong'
        else:
            raise ValueError('Only GRIB files with latlon-projection are '
                             'currently supported, given projection is: %s' %
                             m.projparams)
        lalo = m.latlons()
        x = m.distinctLongitudes
        y = m.distinctLatitudes
        self.xmin = x.min()
        self.xmax = x.max()
        self.ymin = y.min()
        self.ymax = y.max()
        self.delta_x = x[1] - x[0]
        self.delta_y = y[1] - y[0]

        ####################################
        # GRIB source and parameter names
        ####################################
        centre = list(set(centre))
        if len(centre) > 1:
            raise ValueError('File contains data from several centres: ' +
                             str(centre))
        else:
            centre = centre[0]
        if centre in grib_variable_mapping:
            self.grib_mapping = grib_variable_mapping[centre]
            logger.info('Parameter codes not defined in mapper: ' +
                         str(set(marsParams) - set(self.grib_mapping)))
        else:
            raise ValueError(
                'No GRIB variable mapping defined for centre ' + centre)
        self.marsParams = list(set(self.grib_mapping) &
                               set(marsParams))
        self.variables = [self.grib_mapping[v] for v in self.marsParams]

        ####################################
        # Sort by variable and time
        ####################################
        levels = np.array(levels)
        self.indices = {}
        self.levels = {}
        for i, var in enumerate(self.variables):
            m = list(self.grib_mapping.keys())[list(self.grib_mapping.values()).index(var)]
            self.indices[var] = np.where(np.array(marsParams) == m)[0]
            self.levels[var] = levels[self.indices[var]]
        if len(self.variables) > 0:
            self.times = [times[k] for k in self.indices[var]]
            if len(self.times) > 1 and self.times[0] == self.times[1]:  # Duplicate times
                logger.info('Duplicate times for variables, using '
                             'only first occurance.')
                lasttime = None
                for i, tim in enumerate(self.times):
                    if 'newind' not in locals():
                        newind = [i]
                        lasttime = tim
                    else:
                        if tim != lasttime:
                            newind.append(i)
                            lasttime = tim
                newind = np.array(newind, dtype=np.int32)
                self.times = [self.times[i] for i in newind]
                for i, var in enumerate(self.variables):  # Updating indices
                    self.indices[var] = self.indices[var][newind]
                    self.levels[var] = levels[self.indices[var]]
            self.start_time = self.times[0]
            self.end_time = self.times[-1]
            if len(self.times) > 1:
                self.time_step = self.times[-1] - self.times[-2]
            else:
                self.time_step = None
        else:
            self.start_time = None
            self.end_time = None
            self.time_step = None

        # Run constructor of parent Reader class
        super(Reader, self).__init__()

    def get_variables(self, requested_variables, time=None,
                      x=None, y=None, z=None):

        requested_variables, time, x, y, z, outside = self.check_arguments(
            requested_variables, time, x, y, z)

        nearestTime, dummy1, dummy2, indxTime, dummy3, dummy4 = \
            self.nearest_time(time)

        variables = {}
        delta = self.buffer*self.delta_x
        lonmin = np.maximum(x.min() - delta, self.xmin)
        lonmax = np.minimum(x.max() + delta, self.xmax)
        latmin = np.maximum(y.min() - delta, self.ymin)
        latmax = np.minimum(y.max() + delta, self.ymax)

        for var in requested_variables:
            ind = int(self.indices[var][indxTime]) + 1
            msg = self.grib[ind]
            variables[var], lats, lons = msg.data(lat1=latmin, lat2=latmax,
                                                  lon1=lonmin, lon2=lonmax)
        variables['x'] = lons[0, :]
        variables['y'] = lats[:, 0]
        variables['z'] = None
        variables['time'] = nearestTime

        return variables



# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2015,2020 Knut-Frode Dagestad, MET Norway

from opendrift.readers.basereader import BaseReader, StructuredReader


class Reader(BaseReader, StructuredReader):
    '''Reader based on static 2D arrays of variables'''

    def __init__(self, x, y, array_dict, proj4='+proj=latlong'):
        '''init with a map {'variable_name': array, ...}'''

        self.proj4 = proj4
        self.xmin = x.min()
        self.xmax = x.max()
        self.ymin = y.min()
        self.ymax = y.max()

        self.variables = list(array_dict.keys())
        self.array_dict = {'x': x, 'y': y, 'z': 0}
        for var in array_dict:
            self.array_dict[var] = array_dict[var]

        self.start_time = None
        self.end_time = None
        self.time_step = None
        self.name = 'reader_constant_2d'

        # Run constructor of parent Reader class
        super(Reader, self).__init__()

    def get_variables(self, requestedVariables, time=None,
                      x=None, y=None, z=None):

        self.array_dict['time'] = time
        return self.array_dict



"""
Utilities for doing computation with the oil data

For the most part, what these do is extract the actual data in an Oil
object with specified units, etc.

There are also some facilities for working with the data,
such as interpolating, etc.

"""

pass


"""
version_update.py

code for dealing with different versions of the data model

NOTE: this may need some refactoring when it gets more complicated

"""

from .version import Version, VersionError

from .oil import ADIOS_DATA_MODEL_VERSION


class Updater:
    ver_from = None
    ver_to = None

    def __call__(self, py_json):
        this_ver = Version(py_json.get('adios_data_model_version'))
        if this_ver == self.ver_to:  # nothing to be done
            return py_json
        elif this_ver != self.ver_from:  # can't update this
            return py_json
        else:  # we can do the update
            return self.update(py_json)

class update_0_10_0_to_0_10_0(Updater):

    ver_from = Version(0, 10, 0)
    ver_to = Version(0, 11, 0)

    def update(self, py_json):
        # just in case
        this_ver = Version(py_json.get('adios_data_model_version'))

        if this_ver != self.ver_from:
            raise ValueError("Update called with JSON of wrong version"
                             f"JSON version: {this_ver}, this updater "
                             f"can update: {self.ver_from}")
        # change the name of the fraction_weathered attribute
        if 'sub_samples' in py_json:  # very sparse records may not
            for ss in py_json['sub_samples']:
                md = ss['metadata']
                # note: this may add a fraction_weathered = None, but that's OK
                md['fraction_evaporated'] = md.get('fraction_weathered')
                md.pop('fraction_weathered', None)
        py_json['adios_data_model_version'] = str(self.ver_to)

        return py_json


# NOTE: updaters need to be in order
#       not hard when there is only one :-)
UPDATERS = [update_0_10_0_to_0_10_0()]


def update_json(py_json):
    """
    updates JSON for an oil object from an older version to a newer one
    """

    cur_ver = ADIOS_DATA_MODEL_VERSION
    try:
        ver = Version(py_json['adios_data_model_version'])
    except KeyError:
        # assume it's the version from before we added a version
        ver = Version(0, 10, 0)
        py_json['adios_data_model_version'] = str(ver)
    if ver == cur_ver:  # nothing to be done
        return py_json
    # run it through the updaters
    for update in UPDATERS:
        py_json = update(py_json)
    # Now see if it worked
    ver = Version(py_json.get('adios_data_model_version'))
    if ver == cur_ver:
        return py_json
    elif ver > cur_ver:
        raise VersionError(f"Version: {ver} is not supported by this version of Oil object")
    else:
        raise VersionError(f"updater not available for version: {ver}")



# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2015, Knut-Frode Dagestad, MET Norway

import numpy as np
import pyproj

from opendrift.readers.basereader import BaseReader, ContinuousReader


class Reader(BaseReader, ContinuousReader):
    """Artificial reader, with cyclonic surface current around selected centre.

    Purpose is demonstration and testing (unittest).
    """

    def __init__(self, lon=2, lat=66,
                 proj4='+proj=stere +lat_0=90 +lon_0=70 +lat_ts=60 ' +
                       '+units=m +a=6.371e+06 +no_defs'):

        self.fileName = 'ArtificialOceanEddy'
        self.name = 'ArtificialOceanEddy'

        self.proj4 = proj4
        self.proj = pyproj.Proj(self.proj4)

        # Calculate x,y of center of eddy from given lon and lat
        x0, y0 = self.proj(lon, lat)
        width = 600000  # 600 km box
        self.pixelsize = 10000  # Artificial 10 km pixel size
        self.delta_x = self.pixelsize
        self.delta_y = self.pixelsize
        self.x0 = x0
        self.y0 = y0
        self.xmin = x0 - width
        self.xmax = x0 + width
        self.ymin = y0 - width
        self.ymax = y0 + width
        self.start_time = None
        self.end_time = None
        self.time_step = None
        self.variables = ['x_sea_water_velocity', 'y_sea_water_velocity']

        # Run constructor of parent Reader class
        super(Reader, self).__init__()

    def get_variables(self, requestedVariables, time=None,
                      x=None, y=None, z=None):

        requestedVariables, time, x, y, z, outside = self.check_arguments(
            requestedVariables, time, x, y, z)

        variables = {}
        # Construct cyclonic current field
        size = int(int(self.xmax-self.xmin) / self.pixelsize + 1)
        X = x - self.x0
        Y = y - self.y0
        radius = np.sqrt(X*X + Y*Y)
        radius[radius == 0] = 1

        variables['time'] = time
        variables['x'] = x
        variables['y'] = y
        variables['x_sea_water_velocity'] = -Y / radius  # np.ones((size,size))
        variables['y_sea_water_velocity'] = X / radius  # np.zeros((size,size))

        return variables



'''
Add labels to a record -- for those that we can use some automatic criteria for

For a given product type:
    some labels can be determined from API and/or Viscosity ranges.

The criteria follows the ASTM (and others) standards, where we can
'''
from math import inf

from ..product_type import types_to_labels
from ....computation.physical_properties import KinematicViscosity

# # Here are all the Product Types:
# ('Crude Oil NOS',
#  'Tight Oil',
#  'Condensate',
#  'Bitumen Blend',
#  'Refined Product NOS',
#  'Fuel Oil NOS',
#  'Distillate Fuel Oil',
#  'Residual Fuel Oil',
#  'Refinery Intermediate',
#  'Solvent',
#  'Bio-fuel Oil',
#  'Bio-Petro Fuel Oil',
#  'Natural Plant Oil',
#  'Lube Oil',
#  'Dielectric Oil',
#  'Other')

# # These are the current labels that aren't mapped yet:

# 'MDO', 'Vacuum Gas Oil'

# these are the labels with no criteria for density or viscosity
# e.g, if it's a "Crude Oil NOS", it's a 'Crude Oil'
synonyms_for_product_types = {
    'Crude Oil', 'Shale Oil', 'Fracking Oil', 'Fuel Oil', 'Residual Fuel', 'Distillate Fuel',
    'Refined Product', 'Condensate', 'Transformer Oil'
}
# If it's an exact match, then it's definitely a synonym
for pt, labels in types_to_labels.labels.items():
    for label in labels:
        if label == pt:
            # print(f"adding: {label}")
            synonyms_for_product_types.add(label)

# these are labels that are synonymous to other labels
synonyms_for_labels = {
    'Heavy Fuel Oil': ['HFO', 'No. 6 Fuel Oil', 'Bunker C'],
    'Kerosene': ['Jet Fuel'],
    'No. 2 Fuel Oil': ['Diesel', 'Home Heating Oil'],
}

no_criteria = {
    "api_min": -inf,
    "api_max": inf,
    "kvis_min": -inf,
    "kvis_max": inf,
    'kvis_temp': 15
}
label_map = {label: no_criteria for label in synonyms_for_product_types}

# this maps the labels according to API and kinematic viscosity (cSt at given temp in C) ranges.
label_map.update({
    # 'example_label': {"api_min": -inf, "api_max": inf, "kvis_min": -inf, "kvis_max": inf, 'kvis_temp': None},
    'Condensate': {
        "api_min": 50,
        "api_max": inf,
        "kvis_min": -inf,
        "kvis_max": inf,
        'kvis_temp': 15
    },
    'Light Crude': {
        "api_min": 35,
        "api_max": 50,
        "kvis_min": -inf,
        "kvis_max": inf,
        'kvis_temp': 15
    },
    'Medium Crude': {
        "api_min": 20,
        "api_max": 35,
        "kvis_min": -inf,
        "kvis_max": inf,
        'kvis_temp': 15
    },
    'Heavy Crude': {
        "api_min": -inf,
        "api_max": 20,
        "kvis_min": -inf,
        "kvis_max": inf,
        'kvis_temp': 15
    },
    'Group V': {
        "api_min": -inf,
        "api_max": 10.0,
        "kvis_min": -inf,
        "kvis_max": inf,
        'kvis_temp': 15
    },
    'Heavy Fuel Oil': {
        "api_min": -inf,
        "api_max": 15.0,
        "kvis_min": 200,
        "kvis_max": inf,
        'kvis_temp': 50
    },

    # pretty much made this up ... non-newtonian
    'Bitumen': {
        "api_min": -inf,
        "api_max": 10,
        "kvis_min": 1000,
        "kvis_max": inf,
        'kvis_temp': 40
    },
    # I went through all the bitumen records that we have, and most of them do not have
    # kinematic viscosity measurements. However, they all have dynamic viscosity
    # measurements which were > 10^6 cP at 0C and > 10^5 cP at 15C.
    # I propose changing the criteria to reflect these measurements.

    #***** The following need more verification
    # Refined light products
    'No. 2 Fuel Oil': {
        "api_min": 30,
        "api_max": 39,
        "kvis_min": 2.5,
        "kvis_max": 4,
        'kvis_temp': 38
    },
    'Kerosene': {
        "api_min": 47.6,
        "api_max": 67.8,
        "kvis_min": -inf,
        "kvis_max": inf,
        'kvis_temp': 38
    },
    'Aviation Gas': {
        "api_min": 47.6,
        "api_max": 70.8,
        "kvis_min": -inf,
        "kvis_max": inf,
        'kvis_temp': 38
    },
    'Gasoline': {
        "api_min": 59.7,
        "api_max": 76.6,
        "kvis_min": -inf,
        "kvis_max": 2.5,
        'kvis_temp': 38
    },

    # Intermediate Fuel Oils
    'MDO': {
        "api_min": 30,
        "api_max": 42,
        "kvis_min": -inf,
        "kvis_max": 11,
        'kvis_temp': 40
    },
    'IFO': {
        "api_min": 15,
        "api_max": 30,
        "kvis_min": 4,
        "kvis_max": 200,
        'kvis_temp': 38
    },
    #IFO needs an additional limitation to not be tied to biodiesel
})

for label, synonyms in synonyms_for_labels.items():
    label_map.update({syn: label_map[label] for syn in synonyms})


def get_suggested_labels(oil):
    """
    get the labels suggested for this oil

    :param oil: the oil object to get the labels for
    :type oil: Oil object
    """
    labels = set()
    pt = oil.metadata.product_type
    # everything gets its product type as a label as well
    # unless it has no product type
    # if pt:
    #     labels.add(pt)
    if pt == "Other":  # we don't want any labels auto added for Other
        return sorted(labels)
    try:
        for label in types_to_labels.left[oil.metadata.product_type]:
            if is_label(oil, label):
                labels.add(label)
    except KeyError:
        pass

    # sorting so they'll be in consistent order
    return sorted(labels)


def add_labels_to_oil(oil):
    """
    add labels to the passed in oil.

    this adds labels, but does not remove any existing ones

    :param oil: the oil object to add labels to
    :type oil: Oil object
    """
    labels = set(oil.metadata.labels)
    for label in types_to_labels.left['oil.metadata.product_type']:
        if is_label(oil, label):
            labels.add(label)
    oil.metadata.labels = sorted(labels)


def is_label(oil, label):
    try:
        data = label_map[label]
    except KeyError:
        return False

    api = oil.metadata.API

    # check API:
    if ((data['api_min'] != -inf) or (data['api_max'] != inf)):
        if api is None:
            is_label = False
        else:
            is_label = True if data['api_min'] <= api < data['api_max'] else False
    else:
        is_label = True

    if is_label and ((data['kvis_min'] != -inf) or
                     (data['kvis_max'] != inf)):  # check viscosity limits
        try:
            KV = KinematicViscosity(oil)
            kvis = KV.at_temp(temp=data['kvis_temp'], kvis_units='cSt', temp_units='C')
            is_label = True if data['kvis_min'] <= kvis < data['kvis_max'] else False
        except (ZeroDivisionError,
                ValueError):  # if it can't do this, we don't apply the label
            is_label = True

    return is_label


# def link_oil_to_labels(oil):
#     '''
#         Here, we have a single oil and we would like to link it to one or more
#         labels based on its properties.
#     '''
#     try:
#         labels = oil['metadata']['labels']
#     except TypeError:
#         labels = oil.metadata.labels
#     except KeyError:
#         labels = []

#     sample = OilEstimation(oil).get_sample()

#     if sample is None:
#         logger.warn('oil: {} has no fresh sample.  Skipping Categorization.'
#                     .format(oil['oil_id']))
#         return

#     if is_crude_light(oil):
#         labels.extend(['Crude Oil', 'Light Crude'])

#     if is_crude_medium(oil):
#         labels.extend(['Crude Oil', 'Medium Crude'])

#     if is_crude_heavy(oil):
#         labels.extend(['Crude Oil', 'Heavy Crude'])

#     if is_refined_fuel_oil_1(oil, sample):
#         labels.extend(['Refined Product',
#                        'Light',
#                        'No. 1 Fuel Oil',
#                        'Gasoline',
#                        'Kerosene'])

#     if is_refined_fuel_oil_2(oil, sample):
#         labels.extend(['Refined Product',
#                        'No. 2 Fuel Oil',
#                        'Diesel',
#                        'Heating Oil'])

#     if is_refined_ifo(oil, sample):
#         labels.extend(['Refined Product',
#                        'Intermediate',
#                        'Fuel Oil'])

#     if is_refined_fuel_oil_6(oil, sample):
#         labels.extend(['Refined Product',
#                        'Heavy',
#                        'Fuel Oil 6',
#                        'HFO',
#                        'Bunker',
#                        'Group V'])

#     if is_generic(oil):
#         labels.extend(['Other', 'Generic'])

#     if len(labels) == 0:
#         labels.extend(['Other'])

#     try:
#         oil['metadata']['labels'] = labels
#     except TypeError:
#         oil.metadata.labels = labels

# def is_crude(oil):
#     try:
#         return ('product_type' in oil['metadata'] and
#                 oil['metadata']['product_type'] is not None and
#                 oil['metadata']['product_type'].lower() == 'crude')
#     except KeyError:
#         return (hasattr(oil.metadata, 'product_type') and
#                 oil.metadata.product_type is not None and
#                 oil.metadata.product_type.lower() == 'crude')

# def is_refined(oil):
#     try:
#         return ('product_type' in oil['metadata'] and
#                 oil['metadata']['product_type'] is not None and
#                 oil['metadata']['product_type'].lower() == 'refined')
#     except KeyError:
#         return (hasattr(oil.metadata, 'product_type') and
#                 oil.metadata.product_type is not None and
#                 oil.metadata.product_type.lower() == 'refined')

# def api_min(oil, oil_api):
#     api = oil['metadata'].get('API', None)

#     return api is not None and api > oil_api

# def api_max(oil, oil_api):
#     api = oil['metadata'].get('API', None)

#     return api is not None and api < oil_api

# def is_crude_light(oil):
#     return is_crude(oil) and api_min(oil, 31.1)

# def is_crude_medium(oil):
#     return is_crude(oil) and api_max(oil, 31.1) and api_min(oil, 22.3)

# def is_crude_heavy(oil):
#     return is_crude(oil) and api_max(oil, 22.3)

# def is_refined_light_products(oil):
#     '''
#        Category Name:
#        - Light Products
#        Parent:
#        - Refined
#        Sample Oils:
#        - Cooper Basin Light Naphtha
#        - kerosene
#        - JP-4
#        - avgas
#        Density Criteria:
#        - API >= 35
#        Kinematic Viscosity Criteria:
#        - v > 0.0 cSt @ 38 degrees Celcius
#     '''
#     raise NotImplementedError

# def is_refined_fuel_oil_1(oil, sample):
#     '''
#        Category Name:
#        - Fuel oil #1/gasoline/kerosene
#        Sample Oils:
#        - gasoline
#        - kerosene
#        - JP-4
#        - avgas
#        Density Criteria:
#        - API >= 35
#        Kinematic Viscosity Criteria:
#        - v <= 2.5 cSt @ 38 degrees Celcius
#     '''
#     return (is_refined(oil) and
#             api_min(oil, 35.0) and
#             is_within_viscosity_range(sample, kvis_max=2.5))

# def is_refined_fuel_oil_2(oil, sample):
#     '''
#        Category Name:
#        - Fuel oil #2/Diesel/Heating Oil
#        Sample Oils:
#        - Diesel
#        - Heating Oil
#        - No. 2 Distillate
#        Density Criteria:
#        - 30 <= API < 39
#        Kinematic Viscosity Criteria:
#        - 2.5 < v <= 4.0 cSt @ 38 degrees Celcius
#     '''
#     return (is_refined(oil) and
#             api_min(oil, 30.0) and
#             api_max(oil, 39.0) and
#             is_within_viscosity_range(sample, kvis_min=2.5, kvis_max=4.0))

# def is_refined_ifo(oil, sample):
#     '''
#        Category Name:
#        - Intermediate Fuel Oil
#        Sample Oils:
#        - IFO 180
#        - Fuel Oil #4
#        - Marine Diesel
#        Density Criteria:
#        - 15 <= API < 30
#        Kinematic Viscosity Criteria:
#        - 4.0 < v < 200.0 cSt @ 38 degrees Celcius
#     '''
#     return (is_refined(oil) and
#             api_min(oil, 15.0) and
#             api_max(oil, 30.0) and
#             is_within_viscosity_range(sample, kvis_min=4.0, kvis_max=200.0))

# def is_refined_fuel_oil_6(oil, sample):
#     '''
#        Category Name:
#        - Fuel Oil #6/Bunker/Heavy Fuel Oil/Group V
#        Sample Oils:
#        - Bunker C
#        - Residual Oil
#        Density Criteria:
#        - API < 15
#        Kinematic Viscosity Criteria:
#        - 200.0 <= v cSt @ 50 degrees Celcius
#     '''
#     return (is_refined(oil) and
#             api_max(oil, 15.0) and
#             is_within_viscosity_range(sample, kvis_min=200.0))

# def is_generic(oil):
#     '''
#         Category Name:
#         - Other->Generic
#         Criteria:
#         - Any oils that have been generically generated.  These are found
#           in the OilLibTest data file.  Basically these oils have a name
#           that is prefixed with '*GENERIC'.
#     '''
#     try:
#         ret = oil['metadata'].get('name', None)
#     except AttributeError:
#         ret = oil.metadata.name

#     if ret is not None:
#         return ret.startswith('*GENERIC')
#     else:
#         return None

# def is_within_viscosity_range(oil_sample, kvis_min=None, kvis_max=None):
#     category_temp = 273.15 + 38

#     viscosity = oil_sample.kvis_at_temp(category_temp)

#     if viscosity is None:
#         return False

#     if kvis_min is not None and kvis_max is not None:
#         return (viscosity > kvis_min) and (viscosity <= kvis_max)
#     elif kvis_min is not None:
#         return viscosity > kvis_min
#     elif kvis_max is not None:
#         return viscosity <= kvis_max
#     else:
#         return True

# Data from: https://www.sigma2c.com/16api_gravity.html

#                Specific Gravity  API
# Naphtha light       0.66-0.70
# Naphtha medium      0.70-0.75
# Naphtha heavy       0.75-0.80
# Crude oil           0.80-0.97
# Aviation gasoline   0.70-0.78....47.6--70.6
# Kerosene            0.71-0.79    47.6--67.8
# Gasoline            0.68-0.74    59.7--76.6
# Gas oil             0.78-0.86    33.0--49.9
# Diesel oil          0.82-0.90    25.7--41.1
# Lubricating oil     0.82-0.92
# Fuel oil            0.92-0.99
# Asphaslitc bitumen  1.00-1.10



#!/usr/bin/env python

"""
classes for options to clean up Records
"""

class Cleanup:

    # the ID is so that we can know which cleanup method this is
    # subclasses should define this!
    ID = None

    def __init__(self, oil):
        """
        initialize a Cleanup object with an oil object

        :param oil: the oil object you want to clean up
        :type oil: adios_db.models.oil.oil.OIl
        """
        self.oil = oil

    def cleanup(self, oil, do_it=False):
        """
        run this particular cleanup option

        :param oil: an Oil object to act on

        :param do_it=False: flag to tell the cleanup to do its thing. If False,
                            the method returns a message. If True, the action is
                            taken, and the Oil object is altered.

        :returns: a message of what could be done, or what was done.
        """
        raise NotImplementedError



# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <http://www.gnu.org/licenses/>.
#
#
##########################################################################
# This reader supports native SCHISM output netcdf files and will
# handle interpolation of current velocities and other variables
# in 2D and 3D.
# The interpolation is done using a cKDtree-approach defined
# form the 2D or 3D mesh nodes
#
# Author: Simon Weppe. MetOcean Solution, MetService New Zealand
##########################################################################

##########################################################################
# This version aims to be more consistent with the new Opendrift release,
# and re-uses some of the methods in unstructured.py, and better follow
# the new logics
#
# Note the interpolation approach is closer to what is done in
# structured.py with cached reader "blocks" that are re-used for
# successive interpolation between two model time steps.
# unstructured.py instead of reading data at each interpolation cycle
#
##########################################################################


import logging
logger = logging.getLogger(__name__)

import numpy as np
from datetime import datetime
from netCDF4 import num2date
from scipy.interpolate import LinearNDInterpolator
from scipy.spatial import cKDTree #cython-based KDtree for quick nearest-neighbor search
import pyproj
from opendrift.readers.basereader import BaseReader, UnstructuredReader
from opendrift.readers.basereader.consts import *
import xarray as xr


class Reader(BaseReader,UnstructuredReader):
    """
    A reader for unstructured (irregularily gridded) `SCHISM` files.

    Args:
        :param filename: a single SCHISM netcdf output file, or a pattern of files.
                         The netCDF file can also be an URL to an OPeNDAP server.
        :type filename: string, required.

        :param name: Name of reader
        :type name: string, optional

        :param proj4: PROJ.4 string describing projection of data.
        :type proj4: string, optional

        :param use_3d: switch to use 3d flows (if available)
        :type use_3d: boolean, optional

    .. seealso::

        py:mod:`opendrift.readers.basereader.unstructured`.
    """
    def __init__(self, filename=None, name=None, proj4=None, use_3d = None):

        if filename is None:
            raise ValueError('Need filename as argument to constructor')
        filestr = str(filename)
        if name is None:
            self.name = filestr
        else:
            self.name = name

        # Default interpolation method, see function interpolate_block()
        self.interpolation = 'linearNDFast'
        self.convolve = None  # Convolution kernel or kernel size

        # [name_used_in_schism : equivalent_CF_name used in opendrift]
        schism_mapping = {
            'dahv': 'x_sea_water_velocity',
            'dahv': 'y_sea_water_velocity',
            'hvel': 'x_sea_water_velocity',
            'hvel': 'y_sea_water_velocity',
            'depth': 'sea_floor_depth_below_sea_level',
            'elev' : 'sea_surface_height',
            'temp' : 'sea_water_temperature',
            'salt' : 'sea_water_salinity',
            'zcor' : 'vertical_levels', # time-varying vertical coordinates
            'sigma': 'ocean_s_coordinate',
            'vertical_velocity' : 'upward_sea_water_velocity'}
            # diffusivity
            # viscosity

        self.return_block = True

        try:
            # Open file, check that everything is ok
            logger.info('Opening dataset: ' + filestr)
            if ('*' in filestr) or ('?' in filestr) or ('[' in filestr):
                logger.info('Opening files with MFdataset')
                # self.dataset = MFdataset(filename)
                self.dataset = xr.open_mfdataset(filename,chunks={'time': 1}, decode_times=False)
            else:
                logger.info('Opening file with dataset')
                # self.dataset = dataset(filename, 'r')
                self.dataset = xr.open_dataset(filename,chunks={'time': 1}, decode_times=False)
        except Exception as e:
            raise ValueError(e)

        # Define projection of input data
        if proj4 is not None: #  user has provided a projection apriori
            self.proj4 = proj4
        else:                 # no input assumes latlon
            logger.debug('Lon and lat are 1D arrays, assuming latlong projection: proj4 = ''+proj=latlong''')
            self.proj4 = '+proj=longlat +ellps=WGS84 +datum=WGS84 +no_defs' #'+proj=latlong'

        # check if 3d data is available and if we should use it
        self.use_3d = use_3d
        if self.use_3d is None : # not specified, use 3d data by default (if available)
            if 'hvel' in self.dataset.variables:
                self.use_3d = True
            else:
                self.use_3d = False

        if self.use_3d and 'hvel' not in self.dataset.variables:
            logger.debug('No 3D velocity data in file - cannot find variable ''hvel'' ')
        elif self.use_3d and 'hvel' in self.dataset.variables:
            if 'zcor' in self.dataset.variables: # both hvel and zcor in files - all good
                self.nb_levels = self.dataset.variables['hvel'].shape[2] #hvel dimensions : [time,node,lev,2]
            else:
                logger.debug('No vertical level information present in file ''zcor'' ... stopping')
                raise ValueError('variable ''zcor'' must be present in netcdf file to be able to use 3D currents')

        logger.debug('Finding coordinate variables.')
        # Find x, y and z coordinates
        for var_name in self.dataset.variables:

            if var_name in ['SCHISM_hgrid_face_x','SCHISM_hgrid_face_y','SCHISM_hgrid_edge_x','SCHISM_hgrid_edge_y']:
                # all these variables have the same standard name projection_x_coordinate,projection_y_coordinate
                # we need to only use :
                # SCHISM_hgrid_node_x as projection_x_coordinate
                # SCHISM_hgrid_node_y as projection_y_coordinate
                # which are the nodes (also called vertices)
                continue

            var = self.dataset.variables[var_name]
            attributes = var.attrs
            att_dict = var.attrs
            standard_name = ''
            long_name = ''
            axis = ''
            units = ''
            CoordinateAxisType = ''
            if 'standard_name' in attributes:
                standard_name = att_dict['standard_name']
            if 'long_name' in attributes:
                long_name = att_dict['long_name']
            if 'axis' in attributes:
                axis = att_dict['axis']
            if 'units' in attributes:
                units = att_dict['units']
            if '_CoordinateAxisType' in attributes:
                CoordinateAxisType = att_dict['_CoordinateAxisType']

            if standard_name == 'longitude' or \
                    long_name == 'longitude' or \
                    var_name == 'longitude' or \
                    axis == 'X' or \
                    CoordinateAxisType == 'Lon' or \
                    standard_name == 'projection_x_coordinate':
                self.xname = var_name
                if var.ndim == 2:
                    # When datasets are concatenated by mfdataset(), coordinates vector (1D) may
                    # be tiled to a 2D array of size (time,node), keep only one vector for x,y
                    var = var[0,:]
                # Fix for units; should ideally use udunits package
                if units == 'km':
                    unitfactor = 1000
                else:
                    unitfactor = 1
                var_data = var.values
                x = var_data*unitfactor
                self.unitfactor = unitfactor
                self.numx = var.shape[0]
            if standard_name == 'latitude' or \
                    long_name == 'latitude' or \
                    var_name == 'latitude' or \
                    axis == 'Y' or \
                    CoordinateAxisType == 'Lat' or \
                    standard_name == 'projection_y_coordinate':
                self.yname = var_name
                if var.ndim == 2:
                    var = var[0,:]
                # Fix for units; should ideally use udunits package
                if units == 'km':
                    unitfactor = 1000
                else:
                    unitfactor = 1
                var_data = var.values
                y = var_data*unitfactor
                self.numy = var.shape[0]
            if standard_name == 'depth' or axis == 'Z':
                var_data = var.values
                if 'positive' not in var.ncattrs() or \
                        var.__dict__['positive'] == 'up':
                    self.z = var_data
                else:
                    self.z = -var_data
            if standard_name == 'time' or axis == 'T' or var_name == 'time':
                var_data = var.values
                # Read and store time coverage (of this particular file)
                time = var_data
                time_units = units
                if 'calendar' in var.attrs:
                    calendar = var.attrs['calendar']
                else:
                    calendar = 'standard'
                self.times = num2date(time, time_units, calendar=calendar)

                self.start_time = self.times[0]
                self.end_time = self.times[-1]
                if len(self.times) > 1:
                    self.time_step = self.times[1] - self.times[0]
                else:
                    self.time_step = None

        if 'x' not in locals():
            raise ValueError('Did not find x-coordinate variable')
        if 'y' not in locals():
            raise ValueError('Did not find y-coordinate variable')

        self.x = x
        self.y = y

        if not (self.x>360.).any() and self.use_3d :
            logger.debug('Native (x,y) coordinates are lon/lat - cannot use 3D data, setting use_3d = False')
            self.use_3d = False
            import pdb;pdb.set_trace()
            # the 3D interpolation currently doesnt work if the x,y coordinates in native netcdf files
            # are not cartesian but geographic. If that is the case, when doing 3D interpolation and tree search,
            # the vertical distance unit is meter, while the horizontal distance unit is degrees, which will return
            # erroneous "closest" nodes in ReaderBlockUnstruct,interpolate()

        # Run constructor of parent Reader class
        super(Reader, self).__init__()

        # compute CKDtree of (static) 2D nodes using _build_ckdtree_() from unstructured.py
        logger.debug('Building CKDtree of static 2D nodes for nearest-neighbor search')
        self.reader_KDtree = self._build_ckdtree_(self.x,self.y)

        # build convex hull of points for particle-in-mesh checks using _build_boundary_polygon_() from unstructured.py
        logger.debug('Building convex hull of nodes for particle''s in-mesh checks')
        self.boundary = self._build_boundary_polygon_(self.x,self.y)

        # Find all variables having standard_name
        self.variable_mapping = {}
        for var_name in self.dataset.variables:
            if var_name in [self.xname, self.yname]: #'depth'
                continue  # Skip coordinate variables
            var = self.dataset.variables[var_name]
            attributes = var.attrs
            att_dict = var.attrs

            if var_name in schism_mapping:
                # current velocity variables ['dahv','hvel'] need special treatment because
                # [u,v] components are saved in the same variable i.e.
                # u = nc.variables['hvel'][0,:,:]
                # v = nc.variables['hvel'][1,:,:]
                #
                # this means both x_sea_water_velocity and y_sea_water_velocity
                # will be mapped to same name. Correct data will be then extracted
                # in get_variables()
                #
                if var_name == 'hvel' and self.use_3d : # then use 3d data
                    self.variable_mapping['x_sea_water_velocity'] = str(var_name)
                    self.variable_mapping['y_sea_water_velocity'] = str(var_name)
                elif var_name == 'hvel' and not self.use_3d: # skip
                    pass
                elif var_name == 'dahv' and self.use_3d : # skip
                    pass
                elif var_name == 'dahv' and not self.use_3d: # then use depth-averaged data
                    self.variable_mapping['x_sea_water_velocity'] = str(var_name)
                    self.variable_mapping['y_sea_water_velocity'] = str(var_name)
                else: # standard mapping
                    self.variable_mapping[schism_mapping[var_name]] = \
                        str(var_name)

        self.variables = list(self.variable_mapping.keys())

        self.xmin = self.x.min()
        self.xmax = self.x.max()
        self.ymin = self.y.min()
        self.ymax = self.y.max()

        # Dictionaries to store blocks of data for reuse (buffering)
        self.var_block_before = {}  # Data for last timestep before present
        self.var_block_after = {}   # Data for first timestep after present

    def get_variables(self, requested_variables, time=None,
                      x=None, y=None, z=None, block=False):

        """ The function extracts 'requested_variables' from the native SCHISM files
            which will then be used in _get_variables_interpolated_() to initialise the ReaderBlockUnstruct objects
            used to interpolate data in space and time

            For now the function will extract the entire slice of data of 'requested_variables' at given 'time'

            There is an option to extract only a subset of data around particles clouds to have less data but
            it means we need to recompute the KDtree of the subset nodes every time in ReaderBlockUnstruct.
            (Speed gain to be tested)

        """
        requested_variables, time, x, y, z, outside = \
            self.check_arguments(requested_variables, time, x, y, z)

        nearestTime, dummy1, dummy2, indxTime, dummy3, dummy4 = \
            self.nearest_time(time)

        x = np.atleast_1d(x)
        y = np.atleast_1d(y)

        variables = {'x': self.x, 'y': self.y, 'z': 0.*self.y,
                     'time': nearestTime}

        # extracts the full slices of requested_variables at time indxTime
        for par in requested_variables:
            if par not in ['x_sea_water_velocity','y_sea_water_velocity'] :
                # standard case - for all variables except current velocities
                var = self.dataset.variables[self.variable_mapping[par]]
                if var.ndim == 1:
                    data = var[:] # e.g. depth
                    logger.debug('reading constant data from unstructured reader %s' % (par))
                elif var.ndim == 2:
                    data = var[indxTime,:] # e.g. 2D temperature
                    logger.debug('reading 2D data from unstructured reader %s' % (par))
                elif var.ndim == 3:
                    data = var[indxTime,:,:] # e.g. 3D salt [time,node,lev]
                    logger.debug('reading 3D data from unstructured reader %s' % (par))
                    # convert 3D data matrix to one column array and define corresponding data coordinates [x,y,z]
                    # (+ update variables dictionary with 3d coords if needed)
                    data,variables = self.convert_3d_to_array(indxTime,data,variables)
                else:
                    raise ValueError('Wrong dimension of %s: %i' %
                                     (self.variable_mapping[par], var.ndim))
            else :
                # requested variables are current velocities
                # In SCHISM netcdf filesboth [u,v] components are saved
                # as two different dimensions of the same variable.
                var = self.dataset.variables[self.variable_mapping[par]]
                if var.ndim == 3: # depth-averaged current data 'dahv' defined at each node and time [time,node,2]
                    if par == 'x_sea_water_velocity':
                       data = var[indxTime,:,0]
                    elif par == 'y_sea_water_velocity':
                       data = var[indxTime,:,1]
                    logger.debug('reading 2D velocity data from unstructured reader %s' % (par))

                elif var.ndim == 4: # #3D current data 'hvel' defined at each node, level, and time [time,node,zcor,2]
                    if par == 'x_sea_water_velocity':
                       data = var[indxTime,:,:,0]   #hvel dimensions : [time,node,lev,2]
                    elif par == 'y_sea_water_velocity':
                       data = var[indxTime,:,:,1] #hvel dimensions : [time,node,lev,2]
                    logger.debug('reading 3D velocity data from unstructured reader %s' % (par))

                    # convert 3D data matrix to one column array and define corresponding data coordinates [x,y,z]
                    # (+ update variables dictionary with 3d coords if needed)
                    data,variables = self.convert_3d_to_array(indxTime,data,variables)

            variables[par] = data # save all data slice to dictionary with key 'par'
            variables[par] = np.asarray(variables[par])

        return variables

    def convert_3d_to_array(self,id_time,data,variable_dict):
        '''
        The function reshapes a data matrix of dimensions = [node,vertical_levels] (i.e. data at vertical levels, at given time step)
        into a one-column array and works out corresponding 3d coordinates [lon,lat,z] using the time-varying
        'zcor' variable in SCHISM files (i.e. vertical level positions).

        These 3D coordinates will be used to build the 3D KDtree for data interpolation and will be added to the 'variable_dict'
        which is eventually passed to get_variables_interpolated().

        args:
            -id_time
            -data
            -variable_dict
        out :
            -flattened 'data' array
            -addition of ['x_3d','y_3d','z_3d'] items to variable_dict if needed.
        '''

        try:
            vertical_levels = self.dataset.variables['zcor'][id_time,:,:]
            # depth are negative down consistent with convention used in OpenDrift
            # if using the netCDF4 library, vertical_levels is masked array where "masked" levels are those below seabed  (= 9.9692100e+36)
            # if using the xarray library, vertical_levels is nan for levels are those below seabed
            # convert to masked array to be consistent with what netCDF4 lib returns
            vertical_levels = np.ma.array(vertical_levels, mask = np.isnan(vertical_levels.data))
            data = np.asarray(data)
            # vertical_levels.mask = np.isnan(vertical_levels.data) # masked using nan's when using xarray
        except:
            logger.debug('no vertical level information present in file ''zcor'' ... stopping')
            raise ValueError('variable ''zcor'' must be present in netcdf file to be able to use 3D currents')
        # flatten 3D data
        data = np.ravel(data[~vertical_levels.mask])

        # add corresponding 3D coordinates to the 'variable_dict' which is eventually passed to get_variables_interpolated()
        # Note: these 3d coordinates will be the same for all 3d data (current,salt/temp etc..), and will change at each reader time step
        # They are saved as ['x_3d','y_3d','z_3d'] rather than ['x','y','z'] as it would break things when both 2d and 3d data are requested.
        if 'z_3d' not in variable_dict.keys():
            # now make a long 3-column array [lon,lat,z] [n_nodes x  3] at which 'data' is defined
            # tile lon/lat data so that array shape match vertical_levels shape
            x_tiled = np.tile(self.x,(self.nb_levels,1)).T # dimensions [node,vertical_levels]
            y_tiled = np.tile(self.y,(self.nb_levels,1)).T
            # lon_tiled = np.tile(self.lon,(self.nb_levels,1)).T # dimensions [node,vertical_levels]
            # lat_tiled = np.tile(self.lat,(self.nb_levels,1)).T
            # arrays are tiled so that lon_tiled[0,:] = return same value i.e. same lon/lat for all z levels
            if x_tiled.shape != vertical_levels.shape:
                import pdb;pdb.set_trace()
            # convert to masked array consistent with vertical_levels
            x_tiled_ma = np.ma.array(x_tiled, mask = vertical_levels.mask)
            y_tiled_ma = np.ma.array(y_tiled, mask = vertical_levels.mask)
            # flatten arrays and add to dictionary
            variable_dict['x_3d'] = np.ravel(x_tiled_ma[~vertical_levels.mask])
            variable_dict['y_3d'] = np.ravel(y_tiled_ma[~vertical_levels.mask])
            variable_dict['z_3d'] = np.ravel(vertical_levels[~vertical_levels.mask])

        return data,variable_dict

    def _get_variables_interpolated_(self, variables, profiles,
                                   profiles_depth, time,
                                   reader_x, reader_y, z):
        """
        This method _must_ be implemented by every reader. Usually by
        subclassing one of the reader types (e.g.
        :class:`structured.StructuredReader`).

        Arguments are in _native projection_ of reader.

        .. seealso:

            * :meth:`get_variables_interpolated_xy`.
            * :meth:`get_variables_interpolated`.
        """

        """
           Here, this function overloads the _get_variables_interpolated_() methods
           available in unstructured.py.  (which currently doesnt make use of blocks)

           The _get_variables_interpolated_() from structured.py uses regularly gridded
           data "ReaderBlock" extracted from the netcdf files (which may possibly be "cached"
           for speed improvements - see code for more detail).

           This function follows a similar approach but is instead using the native
           high-resolution SCHISM data stored in "ReaderBlockUnstruct" which are used to
           interpolate data in space and time.

           The function returns environment data 'env' interpolated at particle positions [x,y]

        """

        # Find reader time_before/time_after
        time_nearest, time_before, time_after, i1, i2, i3 = \
            self.nearest_time(time)
        logger.debug('Reader time:\n\t\t%s (before)\n\t\t%s (after)' %
                      (time_before, time_after))
        # For variables which are not time dependent, we do not care about time
        static_variables = ['sea_floor_depth_below_sea_level', 'land_binary_mask']

        if time == time_before or all(v in static_variables for v in variables):
            time_after = None

        if profiles is not None:
            # If profiles are requested for any parameters, we
            # add two fake points at the end of array to make sure that the
            # requested block has the depth range required for profiles
            mx = np.append(reader_x, [reader_x[-1], reader_x[-1]])
            my = np.append(reader_y, [reader_y[-1], reader_y[-1]])
            mz = np.append(z, [profiles_depth[0], profiles_depth[1]])
        else:
            mx = reader_x
            my = reader_y
            mz = z

        # start interpolation procedure
        #
        # general idea is to create a  new "ReaderBlockUnstruct" class that will be called instead of
        # the regular "ReaderBlock" as in basereader.py
        # This allows re-using almost the same code as structured.py for the block_before/block_after

        self.timer_start('preparing')
        block_before = block_after = None
        blockvariables_before = variables
        blockvars_before = str(variables)
        blockvariables_after = variables
        blockvars_after = str(variables)
        for blockvars in self.var_block_before:
            if all(v in blockvars for v in variables):
                block_before = self.var_block_before[blockvars]
                blockvariables_before = block_before.data_dict.keys()
                blockvars_before = blockvars
                break
            blockvariables_before = variables
            blockvars_before = str(variables)
        for blockvars in self.var_block_after:
            if all(v in blockvars for v in variables):
                block_after = self.var_block_after[blockvars]
                blockvariables_after = block_after.data_dict.keys()
                blockvars_after = blockvars
                break

        # Swap before- and after-blocks if matching times
        if block_before is not None and block_after is not None:
            if block_before.time != time_before:
                if block_after.time == time_before:
                    block_before = block_after
                    self.var_block_before[blockvars_before] = block_before
            if block_after.time != time_after:
                if block_before.time == time_before:
                    block_after = block_before
                    self.var_block_after[blockvars_after] = block_after

        # Fetch data, if no buffer is available
        if block_before is None or \
                block_before.time != time_before:
            reader_data_dict = \
                 self.__convolve_block__(
                 self.get_variables(blockvariables_before, time_before,
                                    mx, my, mz))
            # now use reader_data_dict to initialize
            # a ReaderBlockUnstruct
            logger.debug('initialize ReaderBlockUnstruct var_block_before')
            self.var_block_before[blockvars_before] = \
                ReaderBlockUnstruct(reader_data_dict,
                    KDtree = self.reader_KDtree,
                    interpolation_horizontal=self.interpolation)
            try:
                len_z = len(self.var_block_before[blockvars_before].z)
            except:
                len_z = 1
            logger.debug(('Fetched env-block (size %ix%ix%i) ' +
                          'for time before (%s)') %
                          (len(self.var_block_before[blockvars_before].x),
                           len(self.var_block_before[blockvars_before].y),
                           len_z, time_before))
            block_before = self.var_block_before[blockvars_before]
        if block_after is None or block_after.time != time_after:
            if time_after is None:
                self.var_block_after[blockvars_after] = \
                    block_before
            else:
                reader_data_dict = self.__convolve_block__(
                    self.get_variables(blockvariables_after, time_after, mx,
                                       my, mz))
                self.timer_start('preparing')
                logger.debug('initialize ReaderBlockUnstruct var_block_after')
                self.var_block_after[blockvars_after] = \
                    ReaderBlockUnstruct(
                        reader_data_dict,
                        KDtree = self.reader_KDtree,
                        interpolation_horizontal=self.interpolation)
                try:
                    len_z = len(self.var_block_after[blockvars_after].z)
                except:
                    len_z = 1

                logger.debug(('Fetched env-block (size %ix%ix%i) ' +
                              'for time after (%s)') %
                              (len(self.var_block_after[blockvars_after].x),
                               len(self.var_block_after[blockvars_after].y),
                               len_z, time_after))
                block_after = self.var_block_after[blockvars_after]

        if (block_before is not None and block_before.covers_positions(
            reader_x, reader_y) is False) or (\
            block_after is not None and block_after.covers_positions(
                reader_x, reader_y) is False):
            import pdb;pdb.set_trace()
            logger.warning('Data block from %s not large enough to '
                            'cover element positions within timestep. '
                            'Buffer size (%s) must be increased.' %
                            (self.name, str(self.buffer)))
        self.timer_end('preparing')

        ############################################################
        # Interpolate before/after blocks onto particles in space
        ############################################################
        self.timer_start('interpolation')
        logger.debug('Interpolating before (%s) in space  (%s)' %
                      (block_before.time, self.interpolation))
        env_before, env_profiles_before = block_before.interpolate(
                reader_x, reader_y, z, variables,
                profiles, profiles_depth)

        if (time_after is not None) and (time_before != time):
            logger.debug('Interpolating after (%s) in space  (%s)' %
                          (block_after.time, self.interpolation))
            env_after, env_profiles_after = block_after.interpolate(
                    reader_x, reader_y, z, variables,
                    profiles, profiles_depth)

        self.timer_end('interpolation')
        #######################
        # Time interpolation
        #######################
        self.timer_start('interpolation_time')
        env_profiles = None
        if (time_after is not None) and (time_before != time) and self.return_block is True:
            weight_after = ((time - time_before).total_seconds() /
                            (time_after - time_before).total_seconds())
            logger.debug(('Interpolating before (%s, weight %.2f) and'
                           '\n\t\t      after (%s, weight %.2f) in time') %
                          (block_before.time, 1 - weight_after,
                           block_after.time, weight_after))
            env = {}
            for var in variables:
                # Weighting together, and masking invalid entries
                env[var] = np.ma.masked_invalid((env_before[var] *
                                                (1 - weight_after) +
                                                env_after[var] * weight_after))

                if var in standard_names.keys():
                    invalid = np.where((env[var] < standard_names[var]['valid_min'])
                               | (env[var] > standard_names[var]['valid_max']))[0]
                    if len(invalid) > 0:
                        logger.warning('Invalid values found for ' + var)
                        logger.warning(env[var][invalid])
                        logger.warning('(allowed range: [%s, %s])' %
                                        (standard_names[var]['valid_min'],
                                         standard_names[var]['valid_max']))
                        logger.warning('Replacing with NaN')
                        env[var][invalid] = np.nan
            # Interpolating vertical profiles in time
            if profiles is not None:
                env_profiles = {}
                logger.info('Interpolating profiles in time')
                # Truncating layers not present both before and after
                numlayers = np.minimum(len(env_profiles_before['z']),
                                       len(env_profiles_after['z']))
                env_profiles['z'] = env_profiles_before['z'][0:numlayers+1]
                for var in env_profiles_before.keys():
                    if var == 'z':
                        continue
                    env_profiles_before[var]=np.atleast_2d(env_profiles_before[var])
                    env_profiles_after[var]=np.atleast_2d(env_profiles_after[var])
                    env_profiles[var] = (
                        env_profiles_before[var][0:numlayers, :] *
                        (1 - weight_after) +
                        env_profiles_after[var][0:numlayers, :]*weight_after)
            else:
                env_profiles = None

        else:
            logger.debug('No time interpolation needed - right on time.')
            env = env_before
            if profiles is not None:
                if 'env_profiles_before' in locals():
                    env_profiles = env_profiles_before
                else:
                    # Copying data from environment to vertical profiles
                    env_profiles = {'z': profiles_depth}
                    for var in profiles:
                        env_profiles[var] = np.ma.array([env[var], env[var]])
        self.timer_end('interpolation_time')
        # the masking, rotation etc.. is done in variables.py get_variables_interpolated_xy()

        return env, env_profiles

    # copied from StructuredReader
    def set_convolution_kernel(self, convolve):
        """Set a convolution kernel or kernel size (of array of ones) used by `get_variables` on read variables."""
        self.convolve = convolve

    def __convolve_block__(self, env):
        """
        Convolve arrays with a kernel, if reader.convolve is set
        """
        if self.convolve is not None:
            from scipy import ndimage
            N = self.convolve
            if isinstance(N, (int, np.integer)):
                kernel = np.ones((N, N))
                kernel = kernel / kernel.sum()
            else:
                kernel = N
            logger.debug('Convolving variables with kernel: %s' % kernel)
            for variable in env:
                if variable in ['x', 'y', 'z', 'time']:
                    pass
                else:
                    if env[variable].ndim == 2:
                        env[variable] = ndimage.convolve(env[variable],
                                                         kernel,
                                                         mode='nearest')
                    elif env[variable].ndim == 3:
                        env[variable] = ndimage.convolve(env[variable],
                                                         kernel[:, :, None],
                                                         mode='nearest')
        return env

    def covers_positions_xy(self, x, y, z=0):
        """
        Check which points are within boundary of mesh.
        Wrapper function of covers_positions() from unstructured.py which is called in
        get_variables_interpolated_xy() function from variables.py
        It returns indices of in-mesh points, and in-mesh point coordinates rather than a boolean array (inside/outside)
        Within get_variables_interpolated_xy() from variables.py, data is queried for these in-mesh points only and the
        full array (incl. out of mesh positions) is re-generated with correct masking
        """
        ind_covered = np.where(self.covers_positions(x, y))[0]
        return ind_covered ,x[ind_covered], y[ind_covered]

    def plot_mesh(self, variable=None, vmin=None, vmax=None,
             filename=None, title=None, buffer=1, lscale='auto',plot_time = None):
        """Plot geographical coverage of reader."""

        #####################################
        # In Dev - plot unstructured mesh
        #
        # To do
        #  - plot a given timestep, and levels for flows
        #  - incorporate that to animation as in basemodel.py etc...
        #
        #####################################

        import matplotlib.pyplot as plt
        from matplotlib.patches import Polygon
        import cartopy.crs as ccrs
        import cartopy.feature as cfeature
        from opendrift_landmask_data import Landmask
        import matplotlib.tri as mtri
        fig = plt.figure()

        if plot_time is None :
            plot_time = self.start_time

        corners = self.xy2lonlat([self.xmin, self.xmin, self.xmax, self.xmax],
                                 [self.ymax, self.ymin, self.ymax, self.ymin])
        lonmin = np.min(corners[0]) - buffer*2
        lonmax = np.max(corners[0]) + buffer*2
        latmin = np.min(corners[1]) - buffer
        latmax = np.max(corners[1]) + buffer
        latspan = latmax - latmin

        # Initialise map
        if latspan < 90:
            # Stereographic projection centred on domain, if small domain
            x0 = (self.xmin + self.xmax) / 2
            y0 = (self.ymin + self.ymax) / 2
            lon0, lat0 = self.xy2lonlat(x0, y0)
            sp = ccrs.Stereographic(central_longitude=lon0, central_latitude=lat0)
            ax = fig.add_subplot(1, 1, 1, projection=sp)
            corners_stere = sp.transform_points(ccrs.PlateCarree(), np.array(corners[0]), np.array(corners[1]))
        else:
            # Global map if reader domain is large
            sp = ccrs.Mercator()
            ax = fig.add_subplot(1, 1, 1, projection=sp)
            #map = Basemap(np.array(corners[0]).min(), -89,
            #              np.array(corners[0]).max(), 89,
            #              resolution='c', projection='cyl')

        # GSHHS coastlines
        f = cfeature.GSHHSFeature(scale=lscale, levels=[1],
                                  facecolor=cfeature.COLORS['land'])
        ax.add_geometries(
            f.intersecting_geometries([lonmin, lonmax, latmin, latmax]),
            ccrs.PlateCarree(),
            facecolor=cfeature.COLORS['land'],
            edgecolor='black')

        gl = ax.gridlines(ccrs.PlateCarree())
        gl.xlabels_top = False

        # Get boundary
        npoints = 10  # points per side
        x = np.array([])
        y = np.array([])
        x = np.concatenate((x, np.linspace(self.xmin, self.xmax, npoints)))
        y = np.concatenate((y, [self.ymin]*npoints))
        x = np.concatenate((x, [self.xmax]*npoints))
        y = np.concatenate((y, np.linspace(self.ymin, self.ymax, npoints)))
        x = np.concatenate((x, np.linspace(self.xmax, self.xmin, npoints)))
        y = np.concatenate((y, [self.ymax]*npoints))
        x = np.concatenate((x, [self.xmin]*npoints))
        y = np.concatenate((y, np.linspace(self.ymax, self.ymin, npoints)))
        # from x/y vectors create a Patch to be added to map
        lon, lat = self.xy2lonlat(x, y)
        lat[lat>89] = 89.
        lat[lat<-89] = -89.
        p = sp.transform_points(ccrs.PlateCarree(), lon, lat)
        xsp = p[:, 0]
        ysp = p[:, 1]

        if variable is None: # generic plot to check extents
            boundary = Polygon(list(zip(xsp, ysp)), alpha=0.5, ec='k', fc='b',
                               zorder=100)
            ax.add_patch(boundary)
            # add nodes
            rx = np.array([self.xmin, self.xmax])
            ry = np.array([self.ymin, self.ymax])
            data = self.get_variables('sea_floor_depth_below_sea_level', plot_time,rx, ry, block=True)
            rlon, rlat = self.xy2lonlat(data['x'], data['y']) # node coordinates
            ax.plot(rlon,rlat, '.',transform=ccrs.PlateCarree())
            # set extents
            buf = (xsp.max()-xsp.min())*.1  # Some whitespace around polygon
            buf = 0
            try:
                ax.set_extent([xsp.min()-buf, xsp.max()+buf, ysp.min()-buf, ysp.max()+buf], crs=sp)
            except:
                pass
        if title is None:
            plt.title(self.name)
        else:
            plt.title(title)
        plt.xlabel('Time coverage: %s to %s' %
                   (self.start_time, self.end_time))

        if variable is not None:
            if len(variable) == 1: # simple scalar variable
                variable = variable[0]
                rx = np.array([self.xmin, self.xmax])
                ry = np.array([self.ymin, self.ymax])
                data = self.get_variables(variable, plot_time,rx, ry, block=True)
                rlon, rlat = self.xy2lonlat(data['x'], data['y']) # node coordinates
                data[variable] = np.ma.masked_invalid(data[variable])
                # ax.plot(rlon,rlat, '.',transform=ccrs.PlateCarree())
                face=self.dataset.variables['SCHISM_hgrid_face_nodes'][:,0:3]-1
                # build triangulation
                triang =mtri.Triangulation(rlon,rlat, triangles=face, mask=None)
                # plot the variable
                ax.tricontourf(triang, data[variable],vmin=vmin, vmax=vmax, transform=ccrs.PlateCarree())
                # add the triangular mesh - too slow
                # ax.triplot(triang, transform=ccrs.PlateCarree())

            elif len(variable)>1 and all('water' in var for var in variable):  # vector field variable = ['x_sea_water_velocity','y_sea_water_velocity']
                pass # not implemented yet

            ax.set_extent([xsp.min()-.5, xsp.max()+.5, ysp.min()-.5, ysp.max()+.5], crs=sp)

        try:  # Activate figure zooming
            mng = plt.get_current_fig_manager()
            mng.toolbar.zoom()
        except:
            pass

        if filename is not None:
            plt.savefig(filename)
            plt.close()
        else:
            plt.ion()
            plt.show()
            import pdb;pdb.set_trace()

###########################
# ReaderBlockUnstruct class
###########################

# horizontal_interpolation_methods = {
#     'nearest': Nearest2DInterpolator,
#     'ndimage': NDImage2DInterpolator,
#     'linearND': LinearND2DInterpolator,
#     'linearNDFast': Linear2DInterpolator}


# vertical_interpolation_methods = {
#     'nearest': Nearest1DInterpolator,
#     'linear': Linear1DInterpolator}

class ReaderBlockUnstruct():
    """Class to store and interpolate the data from an *unstructured* reader.
       This is the equivalent of ReaderBlock (regular grid) for *unstructured* grids.

       arguments: (in addition to ReaderBlock)

           KDtree : for nearest-neighbor search (initialized using SCHISM nodes in reader's _init_() )
                    This is read from reader object, so that it is not recomputed every time

    """
    logger = logging.getLogger('opendrift')  # using common logger

    def __init__(self, data_dict,
                 KDtree = None,
                 interpolation_horizontal='linearNDFast',
                 interpolation_vertical='linear'):

        # Make pointers to data values, for convenience
        self.x = data_dict['x']
        self.y = data_dict['y']
        self.time = data_dict['time']
        self.data_dict = data_dict
        del self.data_dict['x']
        del self.data_dict['y']
        del self.data_dict['time']
        try:
            self.z = data_dict['z']
            # probably not valid ..since z are different for each point in SCHISM..rather than fixed
            # This is used for the profile interpolation that is not yet functional
            #
            del self.data_dict['z']
        except:
            self.z = None
        # if some 3d data is provided, save additional dict entries
        if 'z_3d' in data_dict.keys():
            self.x_3d = data_dict['x_3d']
            self.y_3d = data_dict['y_3d']
            self.z_3d = data_dict['z_3d']
            del self.data_dict['x_3d']
            del self.data_dict['y_3d']
            del self.data_dict['z_3d']

        # Initialize KDtree(s)
        # > save the 2D one by default (initizalied during reader __init__()
        # > compute and save the time-varying 3D KDtree if relevant

        logger.debug('saving reader''s 2D (horizontal) KDtree to ReaderBlockUnstruct')
        self.block_KDtree = KDtree # KDtree input to function = one computed during reader's __init__()

        # If we eventually use subset of nodes rather than full mesh, we'll need to re-compute the 2D KDtree
        # as well, instead of re-using the "full" one available from reader's init
        # logger.debug('Compute time-varying KDtree for 2D nearest-neighbor search')
        # self.block_KDtree = cKDTree(np.vstack((self.x,self.y)).T)  # KDtree input to function = one computed during reader's __init__()

        if hasattr(self,'z_3d'):
            # we need to compute a new KDtree for that time step using vertical coordinates at that time step
            logger.debug('Compute time-varying KDtree for 3D nearest-neighbor search (i.e using ''zcor'') ')
            # clean arrays if needed, especially z_3d (get rid of nan's) - keep only non-nan
            #
            # check for nan's
            # if (self.z_3d != self.z_3d).any(): # not required
            #     self.x_3d = self.x_3d[np.where(self.z_3d == self.z_3d)]
            #     self.y_3d = self.y_3d[np.where(self.z_3d == self.z_3d)]
            #     self.z_3d = self.z_3d[np.where(self.z_3d == self.z_3d)]
            # check for infinite values
            if np.isinf(self.z_3d).any() :
                self.z_3d[np.where(np.isinf(self.z_3d))] = 15.0 #limit to +15.0m i.e. above msl

            self.block_KDtree_3d = cKDTree(np.vstack((self.x_3d,self.y_3d,self.z_3d)).T)
            # do we need copy_data=True ..probably not since "data" [self.x_3d,self.y_3d,self.z_3d]
            # will not change without the KDtree being recomputedplt

        # Mask any extremely large values, e.g. if missing netCDF _Fill_value
        filled_variables = set()
        for var in self.data_dict:
            if isinstance(self.data_dict[var], np.ma.core.MaskedArray):
                self.data_dict[var] = np.ma.masked_outside(
                    np.ma.masked_invalid(self.data_dict[var]), -1E+9, 1E+9)
                # Convert masked arrays to numpy arrays
                self.data_dict[var] = np.ma.filled(self.data_dict[var],
                                                   fill_value=np.nan)
            # Fill missing data towards seafloor if 3D
            if isinstance(self.data_dict[var], (list,)):
                logger.warning('Ensemble data currently not extrapolated towards seafloor')
            elif self.data_dict[var].ndim == 3:
                filled = fill_NaN_towards_seafloor(self.data_dict[var])
                if filled is True:
                    filled_variables.add(var)

        if len(filled_variables) > 0:
            logger.debug('Filled NaN-values toward seafloor for :'
                          + str(list(filled_variables)))

        if 'land_binary_mask' in self.data_dict.keys() and \
                interpolation_horizontal != 'nearest':
            logger.debug('Nearest interpolation will be used '
                          'for landmask, and %s for other variables'
                          % interpolation_horizontal)

    def _initialize_interpolator(self, x, y, z=None):
        logger.debug('Initialising interpolator.')
        self.interpolator2d = self.Interpolator2DClass(self.x, self.y, x, y)
        if self.z is not None and len(np.atleast_1d(self.z)) > 1:
            self.interpolator1d = self.Interpolator1DClass(self.z, z)

    def interpolate(self, x, y, z=None, variables=None,
                    profiles=[], profiles_depth=None):
        # Use the KDtree to interpolate data to [x,y,z] particle positions

        # self._initialize_interpolator(x, y, z)

        env_dict = {}
        if profiles is not []:
            profiles_dict = {'z': self.z} # probably not valid...
        for varname, data in self.data_dict.items():
            nearest = False
            # land mask
            if varname == 'land_binary_mask':
                nearest = True
                self.interpolator2d_nearest = Nearest2DInterpolator(self.x, self.y, x, y)
            # ensemble data
            if type(data) is list:
                num_ensembles = len(data)
                logger.debug('Interpolating %i ensembles for %s' % (num_ensembles, varname))
                if data[0].ndim == 2:
                    horizontal = np.zeros(x.shape)*np.nan
                else:
                    horizontal = np.zeros((len(self.z), len(x)))*np.nan
                ensemble_number = np.remainder(range(len(x)), num_ensembles)
                for en in range(num_ensembles):
                    elnum = ensemble_number == en
                    int_full = self._interpolate_horizontal_layers(data[en], nearest=nearest)
                    if int_full.ndim == 1:
                        horizontal[elnum] = int_full[elnum]
                    else:
                        horizontal[:, elnum] = int_full[:, elnum]
            # standard data 2D or 3D
            else:
                # use KDtree to find nearest neighbours and interpolate based on distance, on 2D or 3D
                # https://docs.scipy.org/doc/scipy/reference/generated/scipy.spatial.cKDTree.query.html#scipy.spatial.cKDTree.query
                # print(varname)
                nb_closest_nodes = 3
                DMIN=1.e-10
                if data.shape[0] == self.x.shape[0] : # 2D data- full slice
                    #2D KDtree
                    dist,i=self.block_KDtree.query(np.vstack((x,y)).T,nb_closest_nodes, workers=-1) #quick nearest-neighbor lookup
                    # dist = distance to nodes / i = index of nodes
                elif hasattr(self,'z_3d') and (data.shape[0] == self.x_3d.shape[0]) : #3D data
                    #3D KDtree
                    dist,i=self.block_KDtree_3d.query(np.vstack((x,y,z)).T,nb_closest_nodes, workers=-1) #quick nearest-neighbor lookup
                    # dist = distance to nodes / i = index of nodes
                    ##############################
                    # PLOT CHECKS
                    if False:
                        import matplotlib.pyplot as plt
                        fig = plt.figure()
                        ax = fig.add_subplot(111, projection='3d')
                        ax.scatter(self.x_3d[i[0]].data.tolist(),self.y_3d[i[0]].data.tolist(),self.z_3d[i[0]].data.tolist(),c='r', marker='o')
                        ax.scatter(x[0:1],y[0:1],z[0:1],c='g', marker='o')
                        plt.ion()
                        plt.show()
                    ##############################3

                dist[dist<DMIN]=DMIN
                fac=(1./dist)
                data_interpolated = (fac*data.take(i)).sum(-1)/fac.sum(-1)

                # horizontal = self._interpolate_horizontal_layers(data, nearest=nearest)

            if profiles is not None and varname in profiles:
                # not functional yet...
                # need to lookup what actually is expected here
                profiles_dict[varname] = data_interpolated # horizontal

            env_dict[varname] = data_interpolated #horizontal

        if 'z' in profiles_dict:
            profiles_dict['z'] = np.atleast_1d(profiles_dict['z'])

        return env_dict, profiles_dict

    def _interpolate_horizontal_layers(self, data, nearest=False):
        '''Interpolate all layers of 3d (or 2d) array.'''

        if nearest is True:
            interpolator2d = self.interpolator2d_nearest
        else:
            interpolator2d = self.interpolator2d
        if data.ndim == 2:
            return interpolator2d(data)
        if data.ndim == 3:
            num_layers = data.shape[0]
            # Allocate output array
            result = np.ma.empty((num_layers, len(interpolator2d.x)))
            for layer in range(num_layers):
                result[layer, :] = self.interpolator2d(data[layer, :, :])
            return result


    def covers_positions(self, x, y, z=None):
        '''Check if given positions are covered by this reader block.'''

        indices = np.where((x >= self.x.min()) & (x <= self.x.max()) &
                           (y >= self.y.min()) & (y <= self.y.max()))[0]

        if len(indices) != len(x):
            import matplotlib.pyplot as plt
            plt.ion()
            plt.plot(x,y,'r.')
            box = np.array([(self.x.min(),self.y.min()),\
            (self.x.max(),self.y.min()),\
            (self.x.max(),self.y.max()),\
            (self.x.min(),self.y.max()),\
            (self.x.min(),self.y.min())])
            plt.plot(box[:,0],box[:,1],'k--')
            plt.title('Increase buffer distance around particle cloud')
            import pdb;pdb.set_trace()
            plt.close()

        if len(indices) == len(x):
            return True
        else:
            return False



# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2021, Julien Moreau, Plastic@Bay CIC (UK)

import numpy as np
from datetime import datetime, timedelta
import sys, os, logging

logger = logging.getLogger(__name__)
from opendrift.readers.basereader import BaseReader, UnstructuredReader

try:
    from data_manip.formats.selafin import Selafin
except ImportError:
    # Activating PYTEL
    try:
        pytel = os.path.join(os.environ['HOMETEL'], 'scripts', 'python3')
        if pytel not in sys.path:
            logger.warning('adding telemac python scripts to PYTHONPATH: %s' %
                           pytel)
            sys.path.append(pytel)

        from data_manip.formats.selafin import Selafin
    except (KeyError, ImportError):
        logger.error(
            'Telemac python scripts cannot be found. These are distributed together with the Telemac source code. This reader will not work.'
        )


class Reader(BaseReader, UnstructuredReader):
    """
    A reader for unstructured (irregularily gridded) `Telemac3D` files.
    Args:
        :param filename: A single Selafin file
        :type filename: string, required.
        :param name: Name of reader
        :type name: string, optional
        :param proj4: PROJ.4 string describing projection of data.
        :type proj4: string, optional
    .. seealso::
        py:mod:`opendrift.readers.basereader.unstructured`.
    """
    def __init__(self, filename=None, name=None, proj4=None, start_time=None):
        def vardic(vars_slf):
            """
            Match the selafin variables from Telemac 3D to the variables used in
            OpenDrift.
            """
            # Define all the variables used in OpenDrift as a dictionary
            # This is done to the best of our knowledge
            Vars_OD = {
                'VELOCITY U      ': 'x_sea_water_velocity',
                'VELOCITY V      ': 'y_sea_water_velocity',
                'VELOCITY W      ': 'upward_sea_water_velocity',
                'TURBULENT ENERGY': 'turbulent_kinetic_energy',
                'TEMPERATURE     ': 'sea_water_temperature',
                'SALINITY        ': 'sea_water_salinity',
                'NUZ FOR VELOCITY': 'ocean_vertical_diffusivity',
                'NUX FOR VELOCITY': 'horizontal_diffusivity',
                'ELEVATION Z     ': 'sea_floor_depth_below_sea_level',
            }

            No_OD_equiv = {
                'x_wind', 'y_wind', 'wind_speed',
                'sea_floor_depth_below_sea_level', 'wind_from_direction',
                'sea_ice_x_velocity', 'sea_ice_y_velocity',
                'sea_surface_wave_significant_height',
                'sea_surface_wave_stokes_drift_x_velocity',
                'sea_surface_wave_stokes_drift_y_velocity',
                'sea_surface_wave_period_at_variance_spectral_density_maximum',
                'sea_surface_wave_mean_period_from_variance_spectral_density_second_frequency_moment',
                'sea_ice_area_fraction', 'surface_downward_x_stress',
                'surface_downward_y_stress', 'turbulent_generic_length_scale'
            }
            No_Telemac_equiv = {
                'NUY FOR VELOCITY',
                'DISSIPATION     ',
            }
            variables = []
            var_idx = []
            for i, var in enumerate(vars_slf):
                try:
                    variables.append(Vars_OD[var])
                    var_idx.append(i)
                except:
                    logger.info(
                        "Selafin variable {} has no equivalent in OpenDrift".
                        format(var))
            return np.array(variables), np.array(var_idx)

        self.name = name if name is not None else filename

        self.timer_start("open dataset")
        logger.info('Opening dataset: %s' % filename)
        self.slf = Selafin(filename)

        logger.info("File:\n{}\nTitle:\n{}".format(self.slf.file, \
                                self.slf.title))
        logger.info('Using projection: %s.' % proj4)
        self.proj4 = proj4

        logger.info('Reading 2D grid')

        # Run constructor of parent Reader class
        super().__init__()
        self.boundary = self._build_boundary_polygon_( \
                        self.slf.meshx,self.slf.meshy)

        self.timer_start("build index")
        logger.debug("building index of nodes...")
        # using selafin method (scipy)
        self.x, self.y = self.slf.meshx, self.slf.meshy

        logger.debug('nodes: %d' % len(self.x))

        # using scipy directly
        self.tree = self._build_ckdtree_(self.slf.meshx, self.slf.meshy)
        #bounds
        self.xmin,self.ymin,self.xmax, self.ymax= self.slf.meshx.min(),\
                self.slf.meshy.min(), self.slf.meshx.max(),self.slf.meshy.max()

        ### time management
        # Slf files have a wrong start time due to a formating error in Telemac
        if start_time is not None:
            if type(start_time) is datetime:
                self.start_time = start_time
            else:
                logger.warning(
                    "The specified start time is not a datetime object")
        else:
            logger.info("loading the datetime from the selafin file")
            self.start_time = datetime(self.slf.datetime[0],
                                       self.slf.datetime[1],
                                       self.slf.datetime[2],
                                       self.slf.datetime[3],
                                       self.slf.datetime[4])
        self.times = []
        for i in range(len(self.slf.tags['times'])):
            self.times.append(self.start_time +
                              timedelta(seconds=self.slf.tags['times'][i]))
        self.end_time = self.times[-1]
        self.altitude_ID=np.array(self.slf.varindex)[np.array( \
                            self.slf.varnames)=='ELEVATION Z     '].tolist()
        self.meshID=(np.arange(self.slf.nplan)[:,None] \
                     *self.slf.npoin2).astype(int)
        self.variables, self.var_idx = vardic(self.slf.varnames)

        self.timer_end("build index")
        self.timer_end("open dataset")

    def plot_mesh(self):
        """
        Plot the grid mesh. Does not automatically show the figure.
        """
        title = 'Unstructured grid: %s\n%s' % (self.name, self.proj)
        from importlib.util import find_spec
        if find_spec("pyvista") is not None:
            import pyvista as pv
            cells = np.hstack(
                ((np.ones(len(self.slf.ikle2), dtype=np.int) * 3)[:, None],
                 self.slf.ikle2))
            points = np.vstack((self.slf.meshx, self.slf.meshy,
                                np.zeros(len(self.slf.meshx)))).T
            u = pv.PolyData(points, cells)
            plotter = pv.Plotter()
            plotter.add_mesh(u, show_edges=True)
            plotter.show_bounds(mesh=u)
            plotter.view_xy()
            plotter.show(title=title, window_size=[800, 640])
        else:
            import matplotlib.pyplot as plt
            plt.figure()
            plt.scatter(self.slf.meshx,
                        self.slf.meshy,
                        marker='x',
                        color='blue',
                        label='nodes')
            x, y = getattr(self.boundary, 'context').exterior.xy
            plt.plot(x, y, color='green', label='boundary')

            plt.legend()
            plt.title(title)
            plt.xlabel('x [m]')
            plt.ylabel('y [m]')

    def get_variables(self,
                      requested_variables,
                      time=None,
                      x=None,
                      y=None,
                      z=None):
        """
        - Query variables based on the particle coordinates x, y, z
        - find the nearest node in the KD tree
        - extract the z array corresponding.
        - extract the index of the node within the 3D mesh
        - extract the variables at the point
        Args:
            x,y,z: np.arrays(float)
                3D coordinates of the particles
            time: np.datetime64
                age of the particle set
            variables: np.array(int)
                indexes of variables
        Returns:
            variables: dictionary of numpy arrays
        """
        ### nearest time tupple
        frames, duration=self.__nearest_idx__(np.array(self.times).astype( \
                         'datetime64[s]'),np.datetime64(time))
        ### nearest node in 2D
        iii = self.__nearest_ckdtree__(self.tree, x, y)
        # build depth ndarrays of each fibre
        niii = len(iii)
        idx_3D = self.meshID + iii
        # locate the profile dimension
        pm = self.__extractslf__(frames, duration, self.altitude_ID, idx_3D)[0]
        # calculate distance from particles to nearest point altitude
        idx_layer = np.abs(pm - z).argmin(axis=0)
        indices = self.variables == np.array(requested_variables)[:, None]
        Idxs = []
        for line in indices:
            Idxs.append(self.var_idx[line])
        idx_nodes = idx_3D[idx_layer, np.arange(len(idx_layer))]
        vectors = self.__extractslf__(frames, duration, Idxs,
                                      idx_nodes.ravel())
        vars = {}
        for i in range(len(requested_variables)):
            if requested_variables[i] == 'sea_floor_depth_below_sea_level':
                bottom = self.__extractslf__(frames, duration,
                                             self.altitude_ID, idx_3D[0])
                surface = self.__extractslf__(frames, duration,
                                              self.altitude_ID, idx_3D[-1])
                vars['sea_floor_depth_below_sea_level'] = surface - bottom
            else:
                vars[requested_variables[i]] = vectors[i]
        return vars

    @staticmethod
    def __nearest_idx__(array, value):
        """
        we are looking for a tuple describing where the sample is and at which
        distance. So we can calculate the FE solution of the variable value.
        input:
            array: a 1D numpy array
            monotonic array
        output:
            bounds: a tupple
            tupple describing the bounding indexes
            dist = tupple
            the distance between the sample and the index values
            so that dist[0]+dist[1]=0
        """
        distance = (array - value).astype(float)
        nearest = np.argsort(abs(distance))[:2]
        # test exact match and out of bounds
        if (distance == 0).any() | (distance > 0).all() | (distance < 0).all():
            bounds = (nearest[0], None)
            dist = (1, 0)
        else:
            bounds = nearest
            if distance[nearest[0]] == distance[nearest[1]] * -1:
                dist = (.5, .5)
            else:
                prop= abs(distance[nearest[0]])/ \
                   (abs(distance[nearest[0]])+abs(distance[nearest[1]]))
                dist = (1 - prop, prop)
        return bounds, dist

    def __extractslf__(self, frames, duration, index_var, index_nodes):
        """
        extract variables from slf files
        index_var must be a list of integer
        """
        vector1 = self.slf.get_variables_at(frames[0], index_var)[:,
                                                                  index_nodes]
        if frames[1] is not None:
            vector2 = self.slf.get_variables_at(frames[1],
                                                index_var)[:, index_nodes]
        else:
            vector2 = 0

        vector = duration[0] * vector1 + duration[1] * vector2
        vector1, vector2 = None, None
        return vector



"""
Main class that represents an oil record.

This maps to the JSON used in the DB

Having a Python class makes it easier to write importing, validating etc, code.
"""
import copy
import json

from dataclasses import dataclass, field

import unit_conversion as uc

from ..common.utilities import dataclass_to_json

from ...computation.gnome_oil import make_gnome_oil
from ...computation import physical_properties

from .metadata import MetaData
from .sample import SampleList
from .version import Version
from .review_status import ReviewStatus

ADIOS_DATA_MODEL_VERSION = Version(0, 11, 0)

from .version_update import update_json  # noqa: E402

# from .validation.warnings import WARNINGS
from .validation.errors import ERRORS  # noqa: E402


@dataclass_to_json
@dataclass
class Oil:
    oil_id: str  # required
    adios_data_model_version: Version = ADIOS_DATA_MODEL_VERSION
    metadata: MetaData = field(default_factory=MetaData)
    sub_samples: SampleList = field(default_factory=SampleList)
    status: list = field(default_factory=list)
    permanent_warnings: list = field(default_factory=list)
    extra_data: dict = field(default_factory=dict)
    review_status: ReviewStatus = field(default_factory=ReviewStatus)

    def __post_init__(self):
        '''
            Put any validation code here (__init__() is auto-generated by the
            dataclass decorator, and it will clobber any attempt to overload
            the __init__.)
        '''
        if self.oil_id == "":
            raise TypeError("You must supply a non-empty oil_id")
        elif not isinstance(self.oil_id, str):
            raise ValueError("oil_id must be a string")
        # arbitrary limit to catch ridiculous ones (UUIDs are 36 chars)
        self._validate_id(self.oil_id)

    def __str__(self):
        """
        need a custom str here, so we don't get a huge dump of the entire tree
        of data
        """
        return (f"{self.metadata.name}\n"
                f"ID: {self.oil_id}\n"
                f"Product Type: {self.metadata.product_type}")

    @staticmethod
    def _pre_from_py_json(py_json):
        # update the JSON version
        py_json = update_json(py_json)

        # this all now done in the update_json method
        # ver = py_json.get('adios_data_model_version')
        # if Version.from_py_json(ver) != ADIOS_DATA_MODEL_VERSION:
        #     raise ValueError("Can't load this version of the data model")
        # py_json.pop('adios_data_model_version', None)
        return py_json

    @classmethod
    def from_file(cls, infile):
        """
        load an Oil object from the passed in JSON file

        it can be either a path or an open file object

        NOTE: this could be in the decorator -- but we only really need it
              for a full record.
        """
        try:
            py_json = json.load(infile)
        except AttributeError:
            # must not be an open file-like object
            py_json = json.load(open(infile, encoding='utf-8'))

        return cls.from_py_json(py_json)

    @staticmethod
    def _validate_id(id):
        if id == "":
            raise TypeError("You must supply a non-empty oil_id")
        elif not isinstance(id, str):
            raise ValueError("oil_id must be a string")
        # arbitrary limit to catch ridiculous onesL UUID is  36 characters
        elif len(id) > 40:
            raise ValueError("oil_id must be a string less than 40 characters "
                             "in length")

    def validate(self):
        """
        validation specific to the Oil object itself

        validation of sub-objects is automatically applied
        """
        # See if it can be used as a GNOME oil
        # NOTE: This is an odd one, as it puts the information in a different
        #       place
        try:
            # Make a copy, as make_gnome_oil might change it in place.
            make_gnome_oil(copy.deepcopy(self))
            self.metadata.gnome_suitable = True
        except Exception:  # If it barfs for any reason it's not suitable
            self.metadata.gnome_suitable = False

        msgs = []

        # Validate ID
        try:
            self._validate_id(self.oil_id)
        except ValueError:
            msgs.append(ERRORS["E001"].format(self.oil_id))

        # check if API matches density
        API = self.metadata.API
        if API is not None:
            try:
                density_at_60F = (physical_properties
                                  .Density(self).at_temp(60, 'F'))

                calculatedAPI = uc.convert('kg/m^3', 'API', density_at_60F)

                # if not round(API, 1) == round(calculatedAPI, 2):
                # if not isclose(API, calculatedAPI, rel_tol=1e-2):

                if abs(API - round(calculatedAPI, 3)) > 0.2:
                    msgs.append(ERRORS["E043"].format(API, calculatedAPI))
            except (IndexError, ValueError):
                pass

        # always add these:
        msgs.extend("W000: " + m for m in self.permanent_warnings)

        return sorted(set(msgs))

    def reset_validation(self):
        """
        calls the validate method, and updates the status with the result
        """
        self.status = self.validate()

    def to_file(self, outfile, sparse=True):
        """
        save an Oil object as JSON to the passed in file

        it can be either a path or a writable open file object

        NOTE: this could be in the decorator -- but we only really need it
              for a full record.
        """
        try:
            json.dump(self.py_json(sparse=sparse), outfile)
        except AttributeError:
            # must not be an open file-like object
            with open(outfile, 'w', encoding='utf-8') as outfile:
                json.dump(self.py_json(sparse=sparse), outfile, indent=4)

        return None



"""
simple class to hold the review status of a record
"""

from datetime import datetime
from dataclasses import dataclass, field

from ..common.utilities import dataclass_to_json

from ..common.validators import EnumValidator

from .validation.errors import ERRORS
from .validation.warnings import WARNINGS


@dataclass_to_json
@dataclass
class ReviewStatus:
    status: str = "Not Reviewed"
    reviewers: str = ""
    review_date: str = ""
    notes: str = ""

    _status_validator = EnumValidator(["Not Reviewed",
                                       "Under Review",
                                       "Review Complete"],
                                      ERRORS['E013'],
                                      case_insensitive=True)

    def validate(self):
        msgs = []
        if self.review_date:
            try:
                datetime.fromisoformat(self.review_date)
            except ValueError as err:
                msgs.append(WARNINGS["W011"].format("review date", self.review_date, str(err)))

        msgs.extend(self._status_validator(self.status))

        return msgs




# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2019, Knut-Frode Dagestad, MET Norway

import logging; logger = logging.getLogger(__name__)
from opendrift.models.basemodel import OpenDriftSimulation
from opendrift.elements.passivetracer import PassiveTracer


class SeaIceDrift(OpenDriftSimulation):
    """Trajectory model based on the OpenDrift framework.

    Simply propagation with ocean sea ice (this module is not for ice bergs)
    Developed at MET Norway.

    """

    ElementType = PassiveTracer
    required_variables = {
            'sea_ice_x_velocity': {'fallback': None},
            'sea_ice_y_velocity': {'fallback': None},
            'land_binary_mask': {'fallback': None}
        }

    def __init__(self, *args, **kwargs):

        super(SeaIceDrift, self).__init__(*args, **kwargs)


    def update(self):
        """Update positions and properties of elements."""

        # Move particles with sea ice velocity
        self.advect_with_sea_ice()



'''
Generic Measurement Types

These are structures that handle an individual measurment:

They have a value, a unit type and a unit.

They can be converted to other units if need be.

They can accommodate a single value, or a range of values

They can also accommodate a standard deviation and number of replicates.
'''

from dataclasses import dataclass
from math import isclose
import copy

from unit_conversion import convert

from ..common.utilities import dataclass_to_json
from ..common.validators import EnumValidator

# why are these oil specific???
# There should be a project-wide repository for warnings & errors

from ..oil.validation.warnings import WARNINGS
from ..oil.validation.errors import ERRORS

__all__ = [
    'AngularVelocity',
    'Concentration',
    'Density',
    'Dimensionless',
    'DynamicViscosity',
    'InterfacialTension',
    'KinematicViscosity',
    'Length',
    'Mass',
    'MassFraction',
    'VolumeFraction',
    'NeedleAdhesion',
    'Pressure',
    'Temperature',
    'Time',
    'Unitless',
    'AnyUnit',
]

# fixme: why is this here?
# it should be in validation, and the list itself should probably
# be in a data file or something.
# class ProductType(str):
#     _valid_types = ('crude',
#                     'refined',
#                     'bitumen product',
#                     'other')
#     _validator = EnumValidator(_valid_types, WARNINGS["W003"],
#                                case_insensitive=True)

#     def validate(self):
#         return self._validator(self)


@dataclass_to_json
@dataclass
class MeasurementDataclass:
    """
    Data structure to hold a value with a unit

    This accommodates both a single value and a range of values

    There is some complexity here, so everything is optional

    NOTE: another reason for everything to be optional is that
    when working with the web client, empty measurements can be
    created and saved before the values are filled in.

    NOTES:
       If there is a value, there should be no min_value or max_value
       If there is only  a min or max, then it is interpreted as
       greater than or less than

       There needs to be validation on that!

       Fixme: maybe there could be a default unit for each unit type?
    """
    value: float = None
    unit: str = None
    min_value: float = None
    max_value: float = None
    standard_deviation: float = None
    replicates: int = None
    unit_type: str = None

    def __post_init__(self):
        """
        this here so that it can get overidden in subclasses

        it appears dataclasses don't add it to __init__ unless it's here
        """
        pass

    # # We want a less-noisy repr
    # def __repr__(self):
    #     atts = ((att, getattr(self, att)) for att in self.__dataclass_fields__.keys())
    #     atts = (f'{att}={val}' for att, val in atts if val is not None)
    #     return f'{self.__class__.__name__}({", ".join(atts)})'

    # __str__ = __repr__


class MeasurementBase(MeasurementDataclass):
    # need to add these here, so they won't be overwritten by the
    # decorator
    unit_type = None

    def __post_init__(self):
        if self.__class__.unit_type is None:
            raise NotImplementedError("Can't initialize a measurement "
                                      "with no unit_type")
        if self.unit_type is None:
            self.unit_type = self.__class__.unit_type
        self.unit_type = self.unit_type.lower()
        if self.unit_type != self.__class__.unit_type:
            raise ValueError(f"unit_type must be: {self.__class__.unit_type}, "
                             f"not {self.unit_type}")
        super().__post_init__()

    def py_json(self, sparse=True):
        # unit_type is added here, as it's not a settable field
        pj = super().py_json(sparse)
        pj['unit_type'] = self.unit_type

        return pj

    def convert_to(self, new_unit):
        """
        Convert this Measurement object to the specified new unit

        The object is mutated in place.

        If the conversion can not be performed, an Exception will
        be raised, and the object not altered.

        This will also return the object (self) -- but that is a
        deprecated feature -- do not use it!

        If you want a new object, use `converted_to` instead
        """

        new_vals = {
            att: None
            for att in ('value', 'min_value', 'max_value',
                        'standard_deviation')
        }

        for attr in new_vals.keys():
            val = getattr(self, attr)

            if val is not None:
                new_val = convert(self.unit_type, self.unit, new_unit, val)
                new_vals[attr] = new_val

        # if this was all successful
        new_vals['unit'] = new_unit

        self.__dict__.update(new_vals)

        return None

    def converted_to(self, new_unit):
        """
        returns a new Measurement object, converted to the units specified
        """
        new = self.copy()
        new.convert_to(new_unit)
        return new

    def copy(self):
        '''
            There will be cases where we want to be non-destructive, such as
            a function that needs to convert to different units to perform
            calculations, but return the results in the original units.

            And since our convert function does an in-place update, we will
            need a way to preserve the original contents of our dataclass
            before the conversion happens.
        '''
        # fixme: why not use the copy.deepcopy() function here?

        return copy.copy(self)


class Temperature(MeasurementBase):
    unit_type = "temperature"
    fixCK = False  # you can monkey-patch this to turn it on.

    def __post_init__(self):
        super().__post_init__()
        if self.fixCK:
            self.fix_C_K()

    def convert_to(self, new_unit):
        # need to do the "right thing" with standard deviation
        if self.standard_deviation is None:
            # no need for anything special
            super().convert_to(new_unit)
        else:
            new_std = convert("deltatemperature", self.unit, new_unit,
                              self.standard_deviation)
            super().convert_to(new_unit)
            self.standard_deviation = new_std

        return self

    def validate(self):
        msgs = []
        if self is None:  # how can this happen?!?! -- but it does.
            return msgs
        # only do this for C or K
        if self.unit.upper() not in {'C', 'K'}:
            return msgs
        for val in (self.value, self.min_value, self.max_value):
            if val is not None:
                val_in_C = convert(self.unit, "C", val)
                decimal = val_in_C % 1
                if isclose(decimal, 0.15) or isclose(decimal, 0.85):
                    msgs.append(WARNINGS['W010'].format(
                        f"{val:.2f} {self.unit} ({val_in_C:.2f} C)",
                        f"{round(val_in_C):.2f} C"))
        return msgs

    def fix_C_K(self):
        """
        This is a bit of a kludge:

        The correct conversion from C to K is 273.16

        But a lot of data sources (notably the ADIOS2 database)
        used 273.0.

        So we end up with temps like: -0.15 C instead of 0.0 C
        This function will "correct" that.

        Note: it also converts to C
        """
        # only do this if there's a validation issue
        for msg in self.validate():
            if 'W010:' in msg:
                break
        else:  # no issue found
            return

        # convert to C
        self.convert_to('C')
        for attr in ("value", "min_value", "max_value"):
            val = getattr(self, attr)
            if val is not None:
                decimal = val % 1
                if isclose(decimal, 0.15) or isclose(decimal, 0.85):
                    setattr(self, attr, round(val))
        return


class Unitless(MeasurementBase):
    '''
    This is a type for data with no unit at all.
    '''
    unit_type = "unitless"

    def convert_to(self, *args, **kwargs):
        raise TypeError("You can not convert a Unitless measurement")


class Dimensionless(MeasurementBase):
    '''
    This is a type that can be converted to generic fractional amounts,
    but does not refer to a particular measurable quantity.
    '''
    unit_type = "dimensionless"


class Time(MeasurementBase):
    unit_type = "time"


class Length(MeasurementBase):
    unit_type = "length"


class Mass(MeasurementBase):
    unit_type = "mass"


class Concentration(MeasurementBase):
    unit_type = 'concentration'


class MassFraction(MeasurementBase):
    unit_type = "massfraction"
    # add a validator: should be between 0 and 1.0


class VolumeFraction(MeasurementBase):
    unit_type = "volumefraction"
    # add a validator: should be between 0 and 1.0


@dataclass
class MassOrVolumeFraction(MeasurementBase):
    """
    This could be Mass or Volume Fraction, or unspecified.

    Unit_type must be one of:
     - MassFraction
     - VolumeFraction
     - Concentration (could be either mass or volume -- who knows?)

    :param unit_type: the value itself
    :param value: the value itself
    :param unit: the value itself
    :param min_value: the value itself
    :param max_value: the value itself
    :param standard_deviation: the value itself
    :param replicates: the value itself
    :param unit_type: the type of unit -- must be "massfraction", "volumefraction" or "concentration"
    """
    def __init__(self, *args, **kwargs):
        unit_type = kwargs.get("unit_type")
        if unit_type is None:
            raise TypeError("unit_type must be specified")
        try:
            unit_type = unit_type.lower().replace(" ", "")
            if unit_type not in {'massfraction', 'volumefraction', 'concentration'}:
                raise AttributeError
        except AttributeError:
            raise ValueError(
                "unit_type must be one of: 'massfraction', 'volumefraction', 'concentration'\n"
                f"args: {args}, kwargs: {kwargs}")
        kwargs['unit_type'] = unit_type
        super().__init__(*args, **kwargs)

    def __post_init__(self):
        # We don't need the post_init in this case
        pass

    def copy(self):
        '''
        copy an instance

        This needs to be special, to preserve the unit_type attribute
        '''
        # fixme: this should probably be the __copy__ method

        return copy.copy(self)

    def __eq__(self, other):
        """
        So as not to be pedantic with the class -- if the values all match
        """
        return self.__dict__ == other.__dict__


@dataclass
class AnyUnit(MeasurementBase):
    '''
    This is a type for data that could be any unit_type
    '''
    def __init__(self, *args, **kwargs):
        unit_type = kwargs.get("unit_type")
        if unit_type is None:
            raise TypeError("unit_type must be specified")
        try:
            self.unit_type = unit_type.lower().replace(" ", "")
        except AttributeError:
            raise TypeError("unit type must be a valid unit_type string")
        kwargs['unit_type'] = self.unit_type
        super().__init__(*args, **kwargs)

    def __post_init__(self):
        '''
        We don't need the post_init in this case

        overriding it to disable it
        '''
        pass

    def __eq__(self, other):
        """
        So as not to be pedantic with the class -- if the values all match
        """
        return self.__dict__ == other.__dict__


class Density(MeasurementBase):
    unit_type = "density"


class DynamicViscosity(MeasurementBase):
    unit_type = "dynamicviscosity"


class KinematicViscosity(MeasurementBase):
    unit_type = "kinematicviscosity"


class Pressure(MeasurementBase):
    unit_type = "pressure"


class NeedleAdhesion(MeasurementBase):
    unit_type = "needleadhesion"


class InterfacialTension(MeasurementBase):
    unit_type = "interfacialtension"


class AngularVelocity(MeasurementBase):
    unit_type = 'angularvelocity'



# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2019, Ole Baadshaug, MET Norway & Ron Saper, Carleton University Canada.
#
# Caveat: This copyright will not interfere with the open nature of OpenDrift and OpenBerg

import numpy as np
from datetime import timedelta, datetime
import pyproj

from opendrift.readers.basereader import BaseReader, ContinuousReader

class Reader(BaseReader, ContinuousReader):
	variables = ['x_sea_water_velocity', 'y_sea_water_velocity']

	def __init__ (self, obslon, obslat, obstime, obsfile=None,
					wind_east=0, wind_north=0, windreader=None, wind_factor=0.018,
					name=None):
		"""	Reader which statistically extrapolate current forcing.
	 	It uses the residual track obtained by subtracting the wind forcing component
	 	from the past observed motion of a particle.
		"""
		if name is not None:
			self.name = name
		else:
			self.name = 'reader_current_from_observation'

		# Cover whole earth, no validity radius yet
		self.proj4 = '+proj=latlong'
		self.xmin = -180
		self.xmax = 180
		self.ymin = -90
		self.ymax = 90

		self.z = np.ma.array([0.], mask=[False],fill_value=1e+20)


		self.start_time = obstime[1]+timedelta(hours=1)
		self.end_time = self.start_time+timedelta(days=10)
		self.times = np.arange(self.start_time, self.end_time, timedelta(hours=1)).astype(datetime)
		self.time_step = self.times[-1] - self.times[-2]

		time_delta_seconds = (obstime[1]-obstime[0]).total_seconds()


		if windreader is not None:
			x0,y0 = windreader.lonlat2xy(obslon[0], obslat[0])
			x1,y1 = windreader.lonlat2xy(obslon[1], obslat[1])

			#NB! Use wind speeds at observed positions
			wind0 = windreader.get_variables_interpolated_xy(['x_wind', 'y_wind'], x=x0, y=y0, z = self.z, time=obstime[0])[0]
			wind1 = windreader.get_variables_interpolated_xy(['x_wind', 'y_wind'], x=x1, y=y1, z = self.z, time=obstime[1])[0]
			meanwind_x = (wind0['x_wind'][0]+ wind1['x_wind'][0])/2
			meanwind_y = (wind0['y_wind'][0]+ wind1['y_wind'][0])/2
		else:
			meanwind_x = wind_east
			meanwind_y = wind_north # wind_east/north should be given in m/s


		g = pyproj.Geod(ellps='WGS84')
		self.azimuth, backazimuth, self.dist = \
			g.inv(obslon[0], obslat[0], obslon[1], obslat[1], radians=False)

		self.speed = self.dist/(time_delta_seconds)

		#Calculate the average speed in x/y-direction of the residual
		self.x_sea_water_velocity = np.atleast_1d(self.speed*np.sin(np.radians(self.azimuth)) - meanwind_x*wind_factor)
		self.y_sea_water_velocity = np.atleast_1d(self.speed*np.cos(np.radians(self.azimuth)) - meanwind_y*wind_factor)


		super(Reader, self).__init__()

	def get_variables(self, requested_variables, time=None,
                      x=None, y=None, z=None):

		requested_variables, time, x, y, z, outside = self.check_arguments(requested_variables, time, x, y, z)
		nearestTime, dummy1, dummy2, indxTime, dummy3, dummy4 = \
			self.nearest_time(time)

		variables = {}
		variables['x_sea_water_velocity'] = self.x_sea_water_velocity*np.ones(len(x))
		variables['y_sea_water_velocity'] = self.y_sea_water_velocity*np.ones(len(y))

		return variables




# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2015, Knut-Frode Dagestad, MET Norway

from datetime import datetime
import numpy as np

from opendrift.readers.basereader import BaseReader, ContinuousReader


class Reader(BaseReader, ContinuousReader):
    '''Returning values oscillating in time with given amplitude and period'''

    def __init__(self, variable, amplitude, period_seconds=3600*24,
                 phase=0, zero_time=datetime(2017, 1, 1, 0)):
        '''init with '''

        self.variables = [variable]
        self.amplitude = amplitude
        self.period_seconds = period_seconds
        self.zero_time = zero_time
        self.proj4 = '+proj=latlong +datum=WGS84'
        self.xmin = -180
        self.xmax = 180
        self.ymin = -90
        self.ymax = 90
        self.start_time = None
        self.end_time = None
        self.time_step = None
        self.name = 'oscillating_reader'

        # Run constructor of parent Reader class
        super(Reader, self).__init__()

    def get_variables(self, requestedVariables, time=None,
                      x=None, y=None, z=None):

        variables = {'time': time, 'x': x, 'y': y, 'z': z}

        phase = ((time - self.zero_time).total_seconds()/
                 self.period_seconds)*np.pi
        value = self.amplitude*np.sin(phase)
        variables[self.variables[0]] = value*np.ones(x.shape)

        return variables
